# -*- coding: utf-8 -*-
# written by Ralf Biehl at the Forschungszentrum Jülich ,
# Jülich Center for Neutron Science 1 and Institute of Complex Systems 1
#    Jscatter is a program to read, analyse and plot data
#    Copyright (C) 2015-2024  Ralf Biehl
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
r"""
---
Lattice objects describing a lattice of points.

Included are methods to select sublattices as parallelepiped, sphere or side of planes.

The small angle scattering is calculated by js.ff.cloudScattering.

The same method can be used to calculate the wide angle scattering with bragg peaks
using larger scattering vectors to get crystalline bragg peaks of nanoparticles.


**Examples**

A hollow sphere cut to a wedge.
::

 import jscatter as js
 import numpy as np
 grid = js.lattice.scLattice(1/2.,2*8,b=[0])
 grid.inSphere(6,b=1)
 grid.inSphere(4,b=0)
 grid.planeSide([1,1,1],b=0)
 grid.planeSide([1,-1,-1],b=0)
 fig = grid.show()

 q=js.loglist(0.01,5,600)
 ffe=js.ff.cloudScattering(q,grid.points,relError=0.02,rms=0.1)
 p=js.grace()
 p.plot(ffe)
 # fig.savefig(js.examples.imagepath+'/grid_wedge.jpg')

.. image:: ../../examples/images/grid_wedge.jpg
 :align: center
 :height: 300px
 :alt: orientedlatticeStructureFactor asymmetric peaks

A cube decorated with spheres.
::

  import jscatter as js
  import numpy as np
  grid= js.lattice.scLattice(0.2,2*15,b=[0])
  v1=np.r_[4,0,0]
  v2=np.r_[0,4,0]
  v3=np.r_[0,0,4]
  grid.inParallelepiped(v1,v2,v3,b=1)
  grid.inSphere(1,center=[0,0,0],b=2)
  grid.inSphere(1,center=v1,b=3)
  grid.inSphere(1,center=v2,b=4)
  grid.inSphere(1,center=v3,b=5)
  grid.inSphere(1,center=v1+v2,b=6)
  grid.inSphere(1,center=v2+v3,b=7)
  grid.inSphere(1,center=v3+v1,b=8)
  grid.inSphere(1,center=v3+v2+v1,b=9)
  fig = grid.show()

  q=js.loglist(0.01,5,600)
  ffe=js.ff.cloudScattering(q,grid.points,relError=0.02,rms=0.)
  p=js.grace()
  p.plot(ffe)
  # fig.savefig(js.examples.imagepath+'/grid_cubewithspheres.jpg')

.. image:: ../../examples/images/grid_cubewithspheres.jpg
 :align: center
 :height: 300px
 :alt: grid_cubewith spheres



**An explicit  comparison of sc, bcc and fcc** nanoparticles (takes a while )

The diffuse scattering appears as a consequence of thermal disorder described
by random displacements from parameter rms=0.05 nm.
::

 import jscatter as js
 import numpy as np
 q=js.loglist(0.01,35,1500)
 q=np.r_[js.loglist(0.01,3,200),3:40:800j]
 unitcelllength=1.5
 N=8

 scgrid= js.lattice.scLattice(unitcelllength,N)
 sc=js.ff.cloudScattering(q,scgrid.points,relError=50,rms=0.05)
 bccgrid= js.lattice.bccLattice(unitcelllength,N)
 bcc=js.ff.cloudScattering(q,bccgrid.points,relError=50,rms=0.05)
 fccgrid= js.lattice.fccLattice(unitcelllength,N)
 fcc=js.ff.cloudScattering(q,fccgrid.points,relError=50,rms=0.05)

 p=js.grace(1.5,1)
 # smooth with Gaussian to include instrument resolution
 p.plot(sc.X,js.formel.smooth(sc,10, window='gaussian'),legend='sc')
 p.plot(bcc.X,js.formel.smooth(bcc,10, window='gaussian'),legend='bcc')
 p.plot(fcc.X,js.formel.smooth(fcc,10, window='gaussian'),legend='fcc')

 q=q=js.loglist(1,35,100)
 p.plot(q,(1-np.exp(-q*q*0.05**2))/scgrid.shape[0],li=1,sy=0,le='sc diffusive')
 p.plot(q,(1-np.exp(-q*q*0.05**2))/bccgrid.shape[0],li=2,sy=0,le='bcc diffusive')
 p.plot(q,(1-np.exp(-q*q*0.05**2))/fccgrid.shape[0],li=3,sy=0,le='fcc diffusive')

 p.title('Comparison sc, bcc, fcc lattice for a nano cube')
 p.yaxis(scale='l',label='I(Q)')
 p.xaxis(scale='l',label='Q / A\S-1')
 p.legend(x=0.03,y=0.001,charsize=1.5)
 p.text('cube formfactor',x=0.02,y=0.05,charsize=1.4)
 p.text('Bragg peaks',x=4,y=0.05,charsize=1.4)
 p.text('diffusive scattering',x=4,y=1e-6,charsize=1.4)
 # p.save(js.examples.imagepath+'/cubicLatticeComparisonExplicit.jpg',size=(1.4,1),dpi=300)

.. image:: ../../examples/images/cubicLatticeComparisonExplicit.jpg
 :align: center
 :height: 300px
 :alt: grid_cubewith spheres

END
"""

import re
import numbers
import numpy as np
from numpy import linalg as la
from scipy.spatial.transform import Rotation

from ..dataarray import dataArray as dA
from ..dataarray import dataList as dL
from .. import mpl

try:
    from .libs import fscatter

    useFortran = True
except ImportError:
    useFortran = False

try:
    import pymatgen.core

    pymatgenfound = True
except ImportError:
    pymatgenfound = False
    pass

from .. import formel

# tolerance for close to zero
_atol = 1e-12


__all__ = ['lattice', 'rhombicLattice', 'bravaisLattice', 'scLattice', 'bccLattice',
            'fccLattice', 'diamondLattice', 'hexLattice', 'hcpLattice', 'pseudoRandomLattice', 'sqLattice',
            'hex2DLattice', 'lamLattice', 'randomLattice', 'honeycombLattice',
            'latticeFromCIF', 'latticeVectorsFromLatticeConstants', 'latticeFromMDA']


class lattice(object):
    isLattice = True

    def __init__(self):
        """
        Create an arbitrary lattice.

        Please use one of the subclasses below for creation.

        pseudorandom, rhombicLattice, bravaisLattice
        scLattice, bccLattice, fccLattice, diamondLattice, hexLattice,
        hcpLattice, sqLattice, hexLattice, lamLattice

        This base class defines methods valid for all subclasses.


        """
        self._select = None

    def __getitem__(self, item):
        return self.points[item]

    def __setitem__(self, item, value):
        if self[item].shape == np.shape(value):
            self.points[item] = value
        else:
            raise TypeError('Wrong shape of given value')

    @property
    def dimension(self):
        return self.points.shape[1] - 1

    @property
    def X(self):
        """X coordinates for b!=0"""
        return self.points[:, 0]

    @property
    def Xall(self):
        """X coordinates """
        return self._points[:, 0]

    @property
    def Y(self):
        """Y coordinates for b!=0"""
        return self.points[:, 1]

    @property
    def Yall(self):
        """Y coordinates """
        return self._points[:, 1]

    @property
    def Z(self):
        """Z coordinates for b!=0"""
        return self.points[:, 2]

    @property
    def Zall(self):
        """Z coordinates"""
        return self._points[:, 2]

    @property
    def XYZ(self):
        """X,Y,Z coordinates array Nx3  for b!=0"""
        return self.points[:, :3]

    @property
    def XYZall(self):
        """X,Y,Z coordinates array Nx3"""
        return self._points[:, :3]

    @property
    def b(self):
        """Scattering length  for points with b!=0"""
        return self.points[:, 3]

    @property
    def ball(self):
        """Scattering length all points"""
        return self._points[:, 3]

    @property
    def shape(self):
        """Shape for points with b!=0"""
        return self.array.shape

    @property
    def shapeall(self):
        """Shape for points with b!=0"""
        return self.array.shape

    @property
    def type(self):
        """Returns type of the lattice"""
        return self._type

    @property
    def array(self):
        """Coordinates and scattering length as array for b!=0"""
        return np.array(self.points)

    def set_b(self, b):
        """
        Set all points to given scattering length.

        Parameters
        ----------
        b : float or array length points

        """
        self._points[:, 3] = b

    def set_bsel(self, b, select):
        """
        Set b of all points (including points with b==0) according to selection.

        To access all points us the properties grid.??all
        which contains all coordinates and b.

        Parameters
        ----------
        b : float or array length _points
            Scattering length
        select : bool array
            Selection array of len(grid.ball)

        Examples
        --------
        ::

         grid.set_ball(1, grid.Xall > 0)

        """
        if select is None:
            self._points[:, 3] = b
        else:
            self._points[select, 3] = b

    @property
    def points(self):
        """
        Points with scattering length !=0

        grid._points contains all Nx[x,y,z,b]

        """
        return self._points[self.nonzerob]

    @property
    def nonzerob(self):
        """
        Mask of nonzero b.
        """
        return np.abs(self._points[:, 3]) > _atol

    def prune(self, select=None):
        """
        Prune lattice to reduced number of points (in place).

        Parameters
        ----------
        select : bool array xN or function
            A bool array of length grid points to keep only points that are true
            or a bool function that is evaluated for each [x,y,z,b] evaluating to True.
            By default all pounts with b=0 are removed.

        Examples
        --------
        ::

         grid.prune(grid.ball>0)  # bool array
         grid.prune(lambda a:a[3]>0)  # function

        """
        if select is None:
            select = self.nonzerob
        if callable(select):
            sel = np.array([select(point) for point in self._points])
            self._points = self._points[sel, :]
        else:
            # try if it was an bool array
            self._points = self._points[select, :]

        # reset last selection
        self._select = None

    def filter(self, funktion):
        """
        Set lattice points scattering length according to a function.

        All points in the lattice are changed for which funktion returns value !=0 (tolerance 1e-12).

        Parameters
        ----------
        funktion : function returning float
            Function to set lattice points scattering length.
            The function is applied with each i point coordinates (array) as input as .points[i,:3].
            The return value is the corresponding scattering length.

        Examples
        --------
        ::

         # To select points inside of a sphere with radius 5 around [1,1,1]:
         from numpy import linalg as la
         sc=js.sf.scLattice(0.9,10)
         sc.set_b(0)
         sc.filter(lambda xyz: 1 if la.norm(xyz-np.r_[1,1,1])<5 else 0)


         # sphere with  increase from center
         from numpy import linalg as la
         sc=js.sf.scLattice(0.9,10)
         sc.set_b(0)
         sc.filter(lambda xyz: 2*(la.norm(xyz)) if la.norm(xyz)<5 else 0)
         fig=sc.show()

        """
        # get float values
        v = np.array([funktion(point) for point in self._points[:, :3]])
        # set for v !=0, don't change others
        select = np.abs(v) > _atol
        self._points[select, 3] = v[select]
        self._select = select

    def centerOfMass(self):
        """
        Center of mass as center of geometry.
        """
        if self.b.shape[0] > 0:
            return self.points[:, :3].mean(axis=0)
        else:
            return [np.nan, np.nan, np.nan]

    def numberOfAtoms(self):
        """
        Number of Atoms
        """
        return self.points.shape[0]

    def move(self, vector):
        """
        Move all points by vector.

        Parameters
        ----------
        vector : list of 3 float or array
            Vector to shift the points.

        """
        self._points[:, :3] = self._points[:, :3] + np.array(vector)

    def inParallelepiped(self, v1, v2, v3, corner=None, b=1, invert=False):
        """
        Set scattering length for points in parallelepiped.

        Parameters
        ----------
        corner : 3x float
            Corner of parallelepiped
        v1,v2,v3 : each 3x float
            Vectors from origin to 3 corners that define the parallelepiped.
        b:  float
            Scattering length for selected points.
        invert : bool
            Invert selection

        Examples
        --------
        ::

         import jscatter as js
         sc=js.sf.scLattice(0.2,10,b=[0])
         sc.inParallelepiped([1,0,0],[0,1,0],[0,0,1],[0,0,0],1)
         sc.show()
         sc=js.sf.scLattice(0.1,30,b=[0])
         sc.inParallelepiped([1,1,0],[0,1,1],[1,0,1],[-1,-1,-1],2)
         sc.show()


        """
        if corner is None:
            corner = [0., 0., 0.]
        a1 = np.cross(v2, v3)
        b1 = np.cross(v3, v1)
        c1 = np.cross(v1, v2)
        # vectors perpendicular to planes
        a1 = a1 /la.norm(a1)
        b1 = b1 /la.norm(b1)
        c1 = c1 /la.norm(c1)
        da = np.dot(self._points[:, :3] - corner, a1)
        da1 = np.dot(np.array(v1), a1)
        db = np.dot(self._points[:, :3] - corner, b1)
        db1 = np.dot(np.array(v2), b1)
        dc = np.dot(self._points[:, :3] - corner, c1)
        dc1 = np.dot(np.array(v3), c1)
        select = (0 <= da) & (da <= da1) & (0 <= db) & (db <= db1) & (0 <= dc) & (dc <= dc1)
        if invert:
            self._points[~select, 3] = b
        else:
            self._points[select, 3] = b
        self._select = select

    def planeSide(self, vector, center=None, b=1, invert=False):
        """
        Set scattering length for points on one side of a plane.

        Parameters
        ----------
        center : 3x float, default [0,0,0]
            Point in plane.
        vector : list 3x float
            Vector perpendicular to plane.
        b:  float
            Scattering length for selected points.
        invert : bool
            False choose points at origin side. True other side.

        Examples
        --------
        ::

         sc=js.sf.scLattice(1,10,b=[0])
         sc.planeSide([1,1,1],[3,3,3],1)
         sc.show()
         sc.planeSide([-1,-1,0],3)
         sc.show()

        """
        if center is None:
            center = [0, 0, 0]
        v = np.array(vector)
        c = np.array(center)
        v = v / la.norm(v)
        # np.dot is not thread safe
        dd = np.einsum('ij,j', self._points[:, :3] - c, v)
        select = (dd > 0)
        if invert:
            self._points[~select, 3] = b
        else:
            self._points[select, 3] = b
        self._select = select

    def inSphere(self, R, center=None, b=1, invert=False):
        """
        Set scattering length for points in sphere.

        Parameters
        ----------
        center : 3 x float, default [0,0,0]
            Center of the sphere.
        R: float
            Radius of sphere around origin.
        b:  float
            Scattering length for selected points.
        invert : bool
            True to invert selection.

        Examples
        --------
        ::

         import jscatter as js
         sc=js.sf.scLattice(1,15,b=[0])
         sc.inSphere(6,[2,2,2],b=1)
         sc.show()
         sc.inSphere(6,[-2,-2,-2],b=2)
         sc.show()

         sc=js.sf.scLattice(0.8,20,b=[0])
         sc.inSphere(3,[2,2,2],b=1)
         sc.inSphere(3,[-2,-2,-2],b=1)
         sc.show()

         sc=js.sf.scLattice(0.8,20,b=[0])
         sc.inSphere(3,[2,2,2],b=1)
         sc.inSphere(4,[0,0,0],b=2)
         sc.show()


        """
        if center is None:
            center = [0, 0, 0]
        select = la.norm(self._points[:, :3] - np.array(center), axis=1) < abs(R)
        if invert:
            self._points[~select, 3] = b
        else:
            self._points[select, 3] = b
        self._select = select

    def inEllipsoid(self, abc, rotation=None, center=None, b=1, invert=False):
        """
        Set scattering length for points in ellipsoid.

        Parameters
        ----------
        abc: [float,float,float]
            Principal semi axes length of the ellipsoid.
        rotation : scipy.spatial.transform.Rotation object
            Rotation describing the rotation semi axes rotation from the cartesian axes.
            e.g. `scipy.spatial.transform.Rotation.from_rotvec([np.pi/4, 0, 0])`
        center : 3 x float, default [0,0,0]
            Center of the sphere.
        b:  float
            Scattering length for selected points.
        invert : bool
            True to invert selection.

        Examples
        --------
        ::

         import jscatter as js
         from scipy.spatial.transform import Rotation as R
         sc=js.sf.scLattice(0.5,30,b=[0])
         sc.inEllipsoid(abc=[3,1,12],b=1)
         sc.show()
         sc.inEllipsoid(abc=[3,1,12],rotation=R.from_euler('X',45,degrees=True),b=2)
         sc.show()
         sc.inEllipsoid(abc=[3,1,12],rotation=R.from_euler('XY',[45,90],degrees=True),b=3)
         sc.show()



        """
        abc=np.array(abc)
        if center is None:
            center = [0, 0, 0]

        # subtract center
        points = self._points[:, :3] - np.array(center)
        if rotation is not None:
            # this should be a scipy.spatial.transform.rotation object
            points=rotation.apply(points)
        # scale to normal form of ellipsoid and use norm<1
        scaledellipscoord = (points) / abc
        select = la.norm(scaledellipscoord, axis=1) < 1
        if invert:
            self._points[~select, 3] = b
        else:
            self._points[select, 3] = b
        self._select = select

    def inCylinder(self, v, R, a=None, length=0, b=1, invert=False):
        """
        Set scattering length for points within a long cylinder.

        Parameters
        ----------
        a : 3 x float, default [0,0,0]
            Edge point on cylinder axis.
        R : float
            Radius of cylinder.
        v : 3 x float
            Normal vector along cylinder axis.
        length :
            Length along cylinder axis.
             - <0 negative direction
             - >0 positive direction
             - =0 infinite both directions
             - np.inf positive direction, infinite length
             - np.ninf negative direction, infinite length
        b:  float
            Scattering length for selected points.
        invert : bool
            True to invert selection.

        Examples
        --------
        ::

         import jscatter as js
         import numpy as np
         sc=js.sf.scLattice(0.5,30,b=[0])
         sc.inCylinder(v=[1,1,1],a=[5,0,5],R=3)
         sc.inCylinder(v=[1,-1,1],a=[0,5,0],R=2)
         sc.inCylinder(v=[1,0,0],a=[0,-5,-10],length =10,R=2)
         sc.inCylinder(v=[0,1,0],a=[5,0,-10],length =10,R=2)
         sc.inCylinder(v=[1,0,0],a=[-10,-5,10],length =10,R=2,b=2)
         sc.inCylinder(v=[0,1,0],a=[-5,0,10],length =-10,R=2,b=2)
         sc.inCylinder(v=[1,0,0],a=[0,5,-10],length =np.NINF,R=2)
         sc.show()

        """
        if a is None:
            a = [0, 0, 0]
        a = np.array(a)
        n = v / la.norm(v)
        p = self._points[:, :3] - a
        # distance from a  (np.dot is not thread safe)
        pn = np.einsum('ij,j', p, n)
        # select if in distance from axis
        select = la.norm(p - (pn * n[:, None]).T, axis=1) < abs(R)
        # not select if outside cylinder length
        if length > 0:
            select[(pn < 0) | (pn > length)] = False
        elif length < 0:
            select[(pn > 0) | (pn < length)] = False

        if invert:
            self._points[~select, 3] = b
        else:
            self._points[select, 3] = b
        self._select = select

    alongLine = inCylinder

    @property
    def lastSelection(self):
        """
        Give last selection from filter, inSphere, inCylinder, inEllipsoid, planeSide, inParallelepiped.

        """
        return self._select

    def show(self, R=None, cmap='rainbow', fig=None, ax=None, atomsize=1):
        """
        Show the lattice in matplotlib with scattering length color coded.

        Parameters
        ----------
        R : float,None
            Radius around origin to show.
        cmap : colormap
            Colormap. E.g. 'rainbow', 'winter','autumn','gray' to
            color atoms according to their scattering length b.
            Use js.mpl.showColors() for all possibilities.
        fig : matplotlib Figure
            Figure to plot in. If None a new figure is created.
        ax : Axes
            If given this axes is used for plotting.
        atomsize : float
            Sphere size of the atoms.
            Unfortunately matplotlib does not scale the point size when zooming.

        Returns
        -------
         fig handle

        Notes
        -----
        If the three dimensional overlap is wrong this is due to matplotlib.
        matplotlib is not a real 3D graphic program.

        """
        if R is None:
            points = self.points
        else:
            points = self.points[la.norm(self.points[:, :3], axis=1) < R]

        if len(points) == 0:
            raise AttributeError('No points with b>0 to show')
        bmax = points[:, 3].max()
        bmin = points[:, 3].min()
        if fig is None:
            fig = mpl.figure()
        if ax is None:
            ax = fig.add_subplot(1, 1, 1, projection='3d')
        else:
            if isinstance(ax, numbers.Integral):
                if len(fig.axes) == 0:
                    ax = fig.add_subplot(1, ax, ax, projection='3d')
                elif len(fig.axes) > ax:
                    ax = fig.axes[ax - 1]
                    ax.clear()
                else:
                    r, c, i = fig.axes[-1].get_geometry()
                    if r * c < ax:
                        r += 1
                        c += 1
                    ax = fig.add_subplot(r, c, ax, projection='3d')
            else:
                # r,c,i = ax.get_geometry()
                ax.clear()
        # scatter plot of points
        scale = (points[:, 3] - bmin) / bmax
        ax.scatter(points[:, 0], points[:, 1], points[:, 2], c=scale*100,
                   s=atomsize,
                   cmap=cmap, vmin=bmin, vmax=bmax, depthshade=False)

        # try to plot unit cell with edge vectors
        try:
            upoints = np.array(self.unitCellAtomPositions)
            ax.scatter(upoints[:, 0], upoints[:, 1], upoints[:, 2], c=self.unitCellAtoms_b,
                       s=self.unitCellAtoms_b / bmin * atomsize ** 2,
                       cmap=cmap, vmin=bmin, vmax=bmax, depthshade=False)
            # unit cell edge vectors
            lv = self.latticeVectors
            if len(lv) > 1:
                line = [lv[0] * 0, lv[0], lv[0] + lv[1], lv[1], lv[0] * 0]
            if len(lv) > 2:
                line += [lv[2], lv[2] + lv[0], lv[0], lv[2] + lv[0],
                         lv[2] + lv[0] + lv[1], lv[0] + lv[1], lv[2] + lv[0] + lv[1],
                         lv[2] + lv[1], lv[1], lv[2] + lv[1],
                         lv[2] + lv[0] * 0]
            line = np.array(line)
            ax.plot(line[:, 0], line[:, 1], line[:, 2], color='g')
            # lines to unit cell atoms if less than 6
            if len(self.unitCellAtoms) < 7:
                try:
                    for uca in self.unitCellAtomPositions:
                        if la.norm(uca) == 0:
                            # atom in origin
                            pass
                        ax.plot([0, uca[0]], [0, uca[1]], [0, uca[2]], color='b')
                except (AttributeError, IndexError):
                    pass
        except AttributeError:
            pass
        ax.set_xlabel('x axis')
        ax.set_ylabel('y axis')
        ax.set_zlabel('z axis')
        xyzmin = self.XYZ.min()
        xyzmax = self.XYZ.max()
        ax.set_xlim(xyzmin, xyzmax)
        ax.set_ylim(xyzmin, xyzmax)
        ax.set_zlim(xyzmin, xyzmax)
        fig.tight_layout()
        mpl.show(block=False)
        return fig

    def getReciprocalLattice(self, size=2):
        print('Only for rhombic lattices')
        return None

    # noinspection PyUnusedLocal,PyUnusedLocal,PyMethodMayBeStatic
    def rotateGrid2hkl(self, grid, hkl):
        print('Only for rhombic lattices')
        return None

    def rotate(self, axis, angle):
        """
        Rotate points in lattice around axis by angle.

        Parameters
        ----------
        axis : list 3xfloat
            Axis of rotation
        angle :
         Rotation angle in rad



        """
        R = formel.rotationMatrix(axis, angle)
        self._points[:, 3] = np.einsum('ij,kj->ki', R, self._points[:, 3])


class pseudoRandomLattice(lattice):

    def __init__(self, size, numberOfPoints, unitCellAtoms=None, b=None, seed=None):
        """
        Create a  lattice with a pseudo random distribution of points.

        Allows to create 1D, 2D or 3D pseudo random latices.
        The Halton sequence is used with skipping the first seed elements of the Halton sequence.

        Parameters
        ----------
        size :list of 3x float
            Size of the lattice for each dimension relative to origin.
        numberOfPoints : int
            Number of points.
        unitCellAtoms: list of 3x1 array, None=[0,0,0]
            Analog to unit cell atoms but distributed around random point.
            Position vectors vi of atoms in **absolute** units as *random_position+unitCellAtoms[i]*.
            It is not checked if there is an overlap to other atoms.
        b : float,array
            Scattering length of atoms (in unitCellAtoms sequence).
        seed : None, int
            Seed for the Halton sequence by skipping the first seed elements of the sequence.
            If None a random integer between 10 and 1e6 is chosen.

        Returns
        -------
        lattice object
            .array  grid points as numpy array

        Examples
        --------
        ::

         import jscatter as js
         grid=js.sf.pseudoRandomLattice([5,5,5],3000)
         fig=grid.show()

         # three atom basis
         grid=js.sf.pseudoRandomLattice([5,5,5],30,unitCellAtoms=[[0,0,0],[0.1,0.1,0],[0.3,0.3,0]],b=[1,2,3])
         fig=grid.show()


        """
        # super(pseudoRandomLattice, self).__init__()
        super().__init__()
        if unitCellAtoms is None:
            self.unitCellAtoms = [np.r_[0, 0, 0]]
        else:
            self.unitCellAtoms = unitCellAtoms
        if b is None:
            b = 1
        if isinstance(b, numbers.Number):
            b = [b]
        if isinstance(b, numbers.Number):
            b = [b] * len(unitCellAtoms)
        self.unitCellAtoms_b = b
        self._size = size
        if seed is None:
            seed = np.random.randint(10, 1000000)
        self._seed = seed
        self._points = self._makeLattice(numberOfPoints, b, seed)
        self._type = 'pseudorandom'

    def _makeLattice(self, N, pb=None, skip=None):
        dim = np.shape(self._size)[0]
        abc = formel.randomPointsInCube(N, skip=skip, dim=dim)
        abc *= np.array(self._size)
        # append missing dimension and b column
        abc = np.c_[abc, np.zeros((N, 4 - dim))]

        # build lattice with all atoms in full grid with b
        points = np.vstack([abc + np.r_[ev, b] for ev, b in
                                  zip(self.unitCellAtoms, self.unitCellAtoms_b)])
        return points

    def appendPoints(self, N, b=None):
        """
        Add points to pseudorandom lattice. In place.

        The Halton sequence is used with skipping the existing points.

        Parameters
        ----------
        N : int
            Number of points.
        b : float,array
            Scattering length of atoms. If array the sequence is repeated to fill N atoms.

        """
        if b is None:
            b = 1
        if isinstance(b, numbers.Number):
            b = [b]
        newpoints = self._makeLattice(N=N, pb=b, skip=self._seed + self._points.shape[0])
        self._points = np.vstack([self._points, newpoints])
        return


class randomLattice(lattice):

    def __init__(self, size, numberOfPoints, b=None, seed=None):
        """
        Create a  lattice with a random distribution of points.

        Allows to create 1D, 2D or 3D random lattices.
        Uses numpy.random.random.

        Parameters
        ----------
        size :3x float
            Size of the lattice for each dimension relative to origin.
        numberOfPoints : int
            Number of points.
        b : float,array
            Scattering length of atoms. If array the sequence is repeated to fill N atoms.
        seed : None, int
            Seed for random.

        Returns
        -------
        lattice object
            .array  grid points as numpy array

        Examples
        --------
        ::

         import jscatter as js
         grid=js.sf.randomLattice([5,5,5],3000)
         fig=grid.show()

        """
        # super(randomLattice, self).__init__()
        super().__init__()
        self.unitCellAtoms = [np.r_[0, 0, 0]]
        if b is None:
            b = 1
        if isinstance(b, numbers.Number):
            b = [b]
        self.unitCellAtoms_b = b
        self._makeLattice(size, numberOfPoints, b, seed)
        self._type = 'random'

    def _makeLattice(self, size, N, pb=None, seed=None):
        dim = np.shape(size)[0]
        if seed is not None:
            np.random.seed(seed=seed)

        seq = np.random.random((N, dim))
        seq *= np.array(size)
        if dim in [1, 2]:
            seq = np.c_[seq, np.zeros((N, 3 - dim))]
        self._points = np.c_[seq, np.tile(pb, N)[:N]]


def latticeVectorsFromLatticeConstants(A, B, C, a, b, c):
    r"""
    Lattice vectors from lattice constants.

    Parameters
    ----------
    A,B,C : float
        Lattice vector length in units nm.
    a,b,c : float
        Angles between lattice vectors in degrees.
         - :math:`a=\alpha=\measuredangle BC`
         - :math:`a=\beta=\measuredangle AC`
         - :math:`a=\gamma=\measuredangle AB`

    Notes
    -----
    See    `<https://en.wikipedia.org/wiki/Lattice_constant>`_


    """
    aa, bb, cc = np.deg2rad([a, b, c])

    rv1 = A * np.r_[1, 0, 0]
    rv2 = B * np.r_[np.cos(cc), np.sin(cc), 0]
    c1 = np.cos(bb)
    c2 = (np.cos(aa) - np.cos(bb) * np.cos(cc)) / np.sin(cc)
    rv3 = C * np.r_[c1, c2, (1 - c1 ** 2 - c2 ** 2) ** 0.5]
    return rv1, rv2, rv3


# noinspection PyMissingConstructor
class rhombicLattice(lattice):
    isRhombic = True

    def __init__(self, latticeVectors, size, unitCellAtoms=None, b=None):
        """
        Create a rhombic lattice with specified unit cell atoms which can also represent lattices of
        mesoscopic particles.

        Allows to create 1D, 2D or 3D latices by using 1, 2 or 3 latticeVectors.
        This can generate classical Bravais lattices but the unit cell atom may also build
        a coarse grained complex particle like a cylinder or ellipsoid.
        In :py:func:`~.structurefactor.ordered.latticeStructureFactor` the particle formfactor amplitude is taken into account.


        Parameters
        ----------
        latticeVectors : list of array 3x1
            Lattice vectors defining the translation of the unit cell along its principal axes.
        size :list of integer
            A list of integers describing the size in direction of the respective latticeVectors.
            Size is symmetric around zero in interval [-i,..,i] with length 2i+1.
        unitCellAtoms : list of 3x1 array, None=[0,0,0]
            Position vectors vi of atoms in the unit cell in relative units of the lattice vectors [0<x<1].
            For 2D and 1D the unit cell atoms vectors are len(vi)=2 and len(vi)=1.
        b : list of float
            Corresponding scattering length of atoms in the unit cell.

        Returns
        -------
        lattice object
            .array               :  grid points as numpy array
            .unitCellVolume      :   V = a1*a2 x a3 with latticeVectors a1, a2, a3;  if existing.
            .dim                 : dimensionality
            .unitCellAtoms       : Unit cell atoms in relative coordinates
            .unitCellAtoms_b     : Scattering length of specific unit cell atoms

        Examples
        --------
        A diatomic base
        ::

         import jscatter as js
         grid=js.sf.rhombicLattice(latticeVectors=[[1,0,0],[0,1,0],[0,0,1]],
                                    size=[3,3,3],
                                    unitCellAtoms=[[-0.1,-0.1,-0.1],[0.1,0.1,0.1]],
                                    b=[1,2])
         grid.show(1.5)


        A cylinder in a hexagonal lattice
        ::

         import jscatter as js
         import numpy as np

         # create lattice vectors in a simple way
         hex = js.sf.hexLattice(4,8,1)

         # create cylinder with atoms in much shorter distance than hex lattice
         cylinder = js.sf.scLattice(0.3,[6,6,13])
         cylinder.move([0,0,3])
         cylinder.inCylinder(v=[0,0,1], R=0.8, a=[0,0,0.5], length=6, b=0, invert=True)

         # Convert the cylinder points coordinates to
         # fractional unit cell coordinates of the hexagonal lattice
         mat = np.linalg.inv(hex.latticeVectors).T
         cellAtoms = mat @ cylinder.XYZ.T

         grid=js.sf.rhombicLattice(latticeVectors=hex.latticeVectors,
                                    size=[1,1,1],
                                    unitCellAtoms=cellAtoms.T,
                                    b=cylinder.b)
         fig = grid.show()
         # fig.savefig(js.examples.imagepath+'/grid_hexlyotropic.jpg')

        .. image:: ../../examples/images/grid_hexlyotropic.jpg
         :align: center
         :height: 300px
         :alt: orientedlatticeStructureFactor asymetric peaks



        """
        if len(latticeVectors) != len(size):
            raise TypeError('size and latticeVectors not compatible. Check dimension!')
        if unitCellAtoms is None:
            unitCellAtoms = [np.r_[0, 0, 0]]
        if b is None:
            b = 1
        if isinstance(b, numbers.Number):
            b = [b] * len(unitCellAtoms)
        self.unitCellAtoms_b = b
        self.unitCellAtoms = unitCellAtoms
        self.latticeVectors = [np.array(lv) for lv in latticeVectors]
        self.size = np.trunc(size)
        self.dim = len(latticeVectors)
        self._makeLattice()
        self._makeReciprocalVectors()
        self._type = 'rhombic'

    def _makeLattice(self):
        latticeVectors = self.latticeVectors
        size = self.size
        abc = latticeVectors[0] * np.r_[-size[0]:size[0] + 1][:, None]
        if self.dim > 1:
            abc = abc + (latticeVectors[1] * np.r_[-size[1]:size[1] + 1][:, None])[:, None]
        if self.dim > 2:
            abc = abc + (latticeVectors[2] * np.r_[-size[2]:size[2] + 1][:, None, None])[:, None, None]

        # abc are basis atoms positions of all unit cells
        abc = abc.reshape(-1, 3)
        abc = np.c_[(abc, np.zeros(abc.shape[0]))]  # add b

        # build lattice with all atoms in full grid with b
        self._points = np.vstack([abc + np.r_[ev, b] for ev, b in
                                  zip(self.unitCellAtomPositions, self.unitCellAtoms_b)])

    def _makeReciprocalVectors(self):
        """
        Creates the reciprocal vectors

        """
        latticeVectors = self.latticeVectors
        # calc reciprocal vectors
        if len(latticeVectors) == 3:
            V = np.dot(latticeVectors[0], np.cross(latticeVectors[1], latticeVectors[2]))
            self.unitCellVolume = abs(V)
            self.reciprocalVectors = []
            self.reciprocalVectors.append(2 * np.pi / V * np.cross(latticeVectors[1], latticeVectors[2]))
            self.reciprocalVectors.append(2 * np.pi / V * np.cross(latticeVectors[2], latticeVectors[0]))
            self.reciprocalVectors.append(2 * np.pi / V * np.cross(latticeVectors[0], latticeVectors[1]))
        elif len(latticeVectors) == 2:
            R = formel.rotationMatrix(np.r_[0, 0, 1], np.pi / 2)
            self.reciprocalVectors = []
            v0 = np.dot(R, latticeVectors[0])
            v1 = np.dot(R, latticeVectors[1])
            self.unitCellVolume = np.dot(v0, v1)
            self.reciprocalVectors.append(2 * np.pi * v1 / np.dot(latticeVectors[0], v1))
            self.reciprocalVectors.append(2 * np.pi * v0 / np.dot(latticeVectors[1], v0))
        elif len(latticeVectors) == 1:
            R = formel.rotationMatrix(np.r_[0, 0, 1], np.pi / 2)
            v0 = np.dot(R, latticeVectors[0])
            self.unitCellVolume = la.norm(v0)
            self.reciprocalVectors = []
            self.reciprocalVectors.append(2 * np.pi * v0 / la.norm(v0) ** 2)

        return

    @property
    def unitCellAtomPositions(self):
        """
        Absolute positions of unit cell atoms.

        """
        return np.einsum('il,ji', self.latticeVectors, np.array(self.unitCellAtoms)[:, :self.dim])

    def getReciprocalLattice(self, size=2, threshold=1e-3):
        """
        Reciprocal lattice of given size with peak scattering intensity.

        Parameters
        ----------
        size : 3x int or int, default 2
            Number of reciprocal lattice points in each direction (+- direction).
        threshold : float
            Threshold for selection rule as
            select if (f2_hkl > max(f2_hkl)*threshold)

        Returns
        -------
            Array [N x 7] with
             reciprocal lattice vectors                 [:,:3]
             corresponding structure factor fhkl**2>0   [:, 3]
             corresponding indices hkl                  [:,4:]

        Notes
        -----
        The threshold for selection rules allows to exclude forbidden peaks but include these for unit cells
        with not equal scattering length as in these cases the selection rule not applies to full extend.
        This is prefered over a explicit list of selection rules.

        """

        if isinstance(size, numbers.Number):
            size = [size] * 3
        size = np.trunc(size)

        # create hkl lattice
        hkl = np.r_[1, 0, 0] * np.r_[-size[0]:size[0] + 1][:, None]
        if len(self.reciprocalVectors) > 1:
            hkl = hkl + (np.r_[0, 1, 0] * np.r_[-size[1]:size[1] + 1][:, None])[:, None]
        if len(self.reciprocalVectors) > 2:
            hkl = hkl + (np.r_[0, 0, 1] * np.r_[-size[2]:size[2] + 1][:, None, None])[:, None, None]
        hkl = hkl.reshape(-1, 3)

        # reciprocal lattice points
        bbb = hkl[:, :len(self.reciprocalVectors)] @ self.reciprocalVectors

        # calc structure factor
        f2hkl = self._f2hkl(hkl)
        # selection rule
        select = (f2hkl > threshold * f2hkl.max())

        return np.c_[bbb[select], f2hkl[select], hkl[select]]

    def getRadialReciprocalLattice(self, size):
        """
        Get radial distribution of Bragg peaks with unit cell structure factor and multiplicity.

        To get real Bragg peak intensities the dimension in lattice directions has to be included
        (+ Debye-Waller factor and diffusive scattering). Use latticeStructureFactor to include these effects.

        Parameters
        ----------
        size : int
            Size of the lattice as maximum included Miller indices.

        Returns
        -------
            list of [unique q values, structure factor fhkl(q)**2, multiplicity mhkl(q), hkl indices]

        Notes
        -----
        The list contains distinct reflections.
        If multiplicities are different than expected, e.g. 18 for hex2D
        the reason is that peaks are degenerate as  hk {35} and {70} result
        in the same Q adding up the multiplicity 18=6+12 .
        In the radial reciprocal lattice these cannot be separated.

        """
        qxyzb = self.getReciprocalLattice(size)[::-1]  # last to get later positive hkl list
        qxyzb = qxyzb[la.norm(qxyzb[:, 4:], axis=1) <= size]  # only complete radial shells (no corners)
        qr = la.norm(qxyzb[:, :3], axis=1).round(decimals=8)
        f2hkl = qxyzb[:, 3]
        hkl = qxyzb[:, 4:]
        qrunique, qrindex, qrcount = np.unique(qr, return_index=True, return_inverse=False, return_counts=True)

        miller = []
        for qru, qrc in zip(qrunique, qrcount):
            sel = qr == qru
            # take the smallest indices with only positive numbers, the last has the smallest by hkl construction
            hklsel = hkl[sel]  # same qr
            sumhkl = np.sum(hklsel >= 0, axis=1)  # count positive
            mi = np.unique(np.sort(hklsel[sumhkl == sumhkl.max()], axis=1), axis=0)
            miller.append(mi[-1])  # take last

        # q values of unique peaks, scattering strength f2hkl, multiplicity as number of unique peaks from 3D count
        if qrunique[0] > 0:
            return qrunique, f2hkl[qrindex], qrcount, miller
        else:
            return qrunique[1:], f2hkl[qrindex][1:], qrcount[1:], miller[1:]

    def getScatteringAngle(self, wavelength=None, size=13):
        r"""
        Get scattering angle :math:`\theta=2arcsin(q_{hkl}\lambda/4\pi)` in degrees.

        Parameters
        ----------
        wavelength : float, 0.15406
            Wavelength :math:`\lambda` in unit nm, default is  Cu K-alpha 0.15406 nm.
        size : int
            Maximum size of reciprocal lattice.

        Returns
        -------
            array in degrees

        """
        if wavelength is None: wavelength = 0.15406
        qrunique, f2hkl, qrcount, hkl = self.getRadialReciprocalLattice(size=size)
        qw = qrunique * wavelength / 4. / np.pi
        theta = np.rad2deg(2 * np.arcsin(qw[np.abs(qw) < 1]))
        return theta

    def _f2hkl(self, hkl):
        """
        Structure factor f**2_hkl which includes the extinction rules.

        """
        pb = np.array(self.unitCellAtoms_b)[:, None]
        hxkylz = np.einsum('ij,lj', np.array(self.unitCellAtoms)[:, :self.dim], hkl[:, :self.dim])
        fhkl = (pb * np.exp(2j * np.pi * hxkylz)).sum(axis=0)
        return (fhkl * fhkl.conj()).real

    def vectorhkl(self, hkl):
        """
        Get vector corresponding to hkl direction.

        Parameters
        ----------
        hkl : 3x float
            Miller indices

        Returns
        -------
        array 3x float

        """
        # this works for dim 1,2,3
        vhkl = np.sum([i*lV for i, lV in zip(hkl, self.latticeVectors)], axis=0)
        return vhkl

    def rotatePlane2hkl(self, plane, hkl, basis=None):
        """
        Rotate plane points that plane is perpendicular to hkl direction.

        Parameters
        ----------
        plane : array Nx3
            3D points of plane to rotate.
            If None the rotation matrix is returned.
        hkl : list of int, float
            Miller indices as [1,1,1] indicating the lattice direction where to rotate to.
        basis : array min 3x3, default=None
            3 basis points spanning the plane to define rotation.
            e.g.[[0,0,0],[1,0,0],[0,1,0]] for xy plane.
            If None the first points of the plane are used instead.


        Returns
        -------
            rotated plane points array 3xN
            or rotation matrix 3x3

        Notes
        -----
        The rotation matrix may be used to rotate the plane  to the desired direction. ::

         R = grid.rotatePlane2hkl(None,[1,1,1],[[0,0,0],[1,0,0],[0,1,0]] )
         newplanepoints = np.einsum('ij,kj->ki', R, planepoints)

        or the transposed R can be used to rotate the lattice. ::

         R = grid.rotatePlane2hkl(None,[1,1,1],[[0,0,0],[1,0,0],[0,1,0]] )
         grid.rotatebyMatrix(R.T)



        Examples
        --------
        ::

         import jscatter as js
         import numpy as np
         R=8
         N=10
         qxy=np.mgrid[-R:R:N*1j, -R:R:N*1j].reshape(2,-1).T
         qxyz=np.c_[qxy,np.zeros(N**2)]
         fccgrid = js.lattice.fccLattice(2.1, 2)
         xyz=fccgrid.rotatePlane2hkl(qxyz,[1,1,1])

         p=js.mpl.scatter3d(xyz[:,0],xyz[:,1],xyz[:,2])
         p.axes[0].scatter(fccgrid.X,fccgrid.Y,fccgrid.Z)


        """
        if plane is not None:
            plane = np.array(plane)
        if basis is None:
            basis = plane
        else:
            basis = np.array(basis)
        # hkl direction
        vhkl = self.vectorhkl(hkl)
        vhkl = vhkl /la.norm(vhkl)
        # search for vector v3 perpendicular to plane
        # first vector in grid close to point, then next with cross >0
        v1 = basis[1] - basis[0]
        i = 2
        while True:
            v2 = basis[i] - basis[0]
            v3 = np.cross(v1, v2)
            # test if >0 then it is perpendicular to plane as v1not parallel to v2
            if la.norm(v3) > 0:
                break
            else:
                i += 1
        v3 = v3 /la.norm(v3)

        rotvector = np.cross(vhkl, v3)
        if la.norm(rotvector) < 1e-8:
            # is parallel
            if plane is None:
                return np.eye(3)
            else:
                return plane
        else:
            angle = np.arccos(np.clip(np.dot(vhkl / la.norm(vhkl), v3 / la.norm(v3)), -1.0, 1.0))
            R = formel.rotationMatrix(rotvector, angle)
            print(angle,rotvector)
            if plane is None:
                return R
            else:
                return np.einsum('ij,kj->ki', R, plane)

    def rotatePlaneAroundhkl(self, plane, hkl, angle):
        """
        Rotate plane points around hkl direction.

        Parameters
        ----------
        plane : array Nx3, None
            3D points of plane. If None the rotation matrix is returned.
        hkl : list of int, float
            Miller indices as [1,1,1] indicating the lattice direction to rotate around.
        angle : float
            Angle in rad

        Returns
        -------
            plane points array 3xN
        or  rotation matrix


        Notes
        -----
        The rotation matrix may be used to rotate the plane to the desired direction. ::

         R = grid.rotatePlaneAroundhkl(None,[1,1,1],1.234)
         newplanepoints = np.einsum('ij,kj->ki', R, planepoints)

        or the transposed R can be used to rotate the lattice. ::

         grid.rotatebyMatrix(R.T)

        Examples
        --------
        ::

         import jscatter as js
         import numpy as np
         R=8
         N=10
         qxy=np.mgrid[-R:R:N*1j, -R:R:N*1j].reshape(2,-1).T
         qxyz=np.c_[qxy,np.zeros(N**2)]
         fccgrid = js.lattice.fccLattice(2.1, 1)
         xyz=fccgrid.rotatePlane2hkl(qxyz,[1,1,1])
         xyz2=fccgrid.rotatePlaneAroundhkl(xyz,[1,1,1],np.deg2rad(30))
         p=js.mpl.scatter3d(xyz[:,0],xyz[:,1],xyz[:,2])
         p.axes[0].scatter(xyz2[:,0],xyz2[:,1],xyz2[:,2])
         p.axes[0].scatter(fccgrid.X,fccgrid.Y,fccgrid.Z)


        """
        vhkl = self.vectorhkl(hkl)

        R = formel.rotationMatrix(vhkl, angle)
        if plane is None:
            return R
        else:
            Rplane = np.einsum('ij,kj->ki', R, plane)
            return Rplane

    def rotatehkl2Vector(self, hkl, vector):
        """
        Rotate lattice that hkl direction is parallel to vector.

        Includes rotation of latticeVectors.

        Parameters
        ----------
        hkl : 3x float
            Direction given as Miller indices.
        vector : 3x float
            Direction to align to.

        Examples
        --------
        ::

         import jscatter as js
         import numpy as np
         R=8
         N=10
         qxy=np.mgrid[-R:R:N*1j, -R:R:N*1j].reshape(2,-1).T
         qxyz=np.c_[qxy,np.zeros(N**2)]

         fccgrid = js.lattice.fccLattice(2.1, 1)
         p=js.mpl.scatter3d(fccgrid.X,fccgrid.Y,fccgrid.Z)

         fccgrid.rotatehkl2Vector([1,1,1],[1,0,0])
         p.axes[0].scatter(fccgrid.X,fccgrid.Y,fccgrid.Z)

         fccgrid.rotateAroundhkl([1,1,1],np.deg2rad(30))
         p.axes[0].scatter(fccgrid.X,fccgrid.Y,fccgrid.Z)

        """
        vv = np.asarray(vector, dtype=float)
        vv = vv /la.norm(vv)
        vhkl = self.vectorhkl(hkl)
        vhkl = vhkl /la.norm(vhkl)

        rotvector = np.cross(vhkl, vv)
        angle = np.arccos(np.clip(np.dot(vhkl, vv), -1.0, 1.0))
        R = formel.rotationMatrix(rotvector, angle)
        self.rotatebyMatrix(R)

    def rotateAroundhkl(self, hkl, angle=None, vector=None, hkl2=None):
        """
        Rotate lattice around hkl direction by angle or to align to vector.

        Uses angle or aligns hkl2 to vector.
        Includes rotation of latticeVectors.

        Parameters
        ----------
        hkl : 3x float
            Direction given as Miller indices.
        angle : float
            Rotation angle in rad.
        vector : 3x float
            Vector to align hkl2 to. Overrides angle.
            Should not be parallel to hkl direction.
        hkl2 : 3x float
            Direction to align along vector. Overrides angle.
            Should not be parallel to hkl direction.

        Examples
        --------
        ::

         import jscatter as js
         import numpy as np
         R=8
         N=10
         qxy=np.mgrid[-R:R:N*1j, -R:R:N*1j].reshape(2,-1).T
         qxyz=np.c_[qxy,np.zeros(N**2)]
         fccgrid = js.lattice.fccLattice(2.1, 1)
         fig=js.mpl.figure( )
         # create subplot to define geometry
         fig.add_subplot(2,2,1,projection='3d')
         fccgrid.show(fig=fig,ax=1)
         fccgrid.rotatehkl2Vector([1,1,1],[1,0,0])
         fccgrid.show(fig=fig,ax=2)
         fccgrid.rotateAroundhkl([1,1,1],np.deg2rad(30))
         fccgrid.show(fig=fig,ax=3)
         fccgrid.rotateAroundhkl([1,1,1],[1,0,0],[1,0,0])
         fccgrid.show(fig=fig,ax=4)

        """
        vhkl = self.vectorhkl(hkl)
        vhkl = vhkl /la.norm(vhkl)

        if vector is not None and hkl2 is not None:
            vv = np.asarray(vector)
            vv = vv /la.norm(vv)
            vhkl2 = self.vectorhkl(hkl2)
            vhkl2 = vhkl2 /la.norm(vhkl2)
            if np.cross(vhkl, vhkl2) < 1e-8 or np.cross(vhkl, vv) < 1e-8:
                # parallel to hkl
                raise Exception('vector or hkl2 parallel to hkl')
            else:
                angle = np.arccos(np.clip(np.dot(vhkl, vv), -1.0, 1.0))

        R = formel.rotationMatrix(vhkl, angle)
        self.rotatebyMatrix(R)

    def rotatebyMatrix(self, R):
        """
        Rotate lattice by rotation matrix including reciprocal vectors (in place).

        Parameters
        ----------
        R : 3x3 array
            Rotation matrix.

        """
        # test if it is rotation matrix
        if not (np.allclose(la.det(R), 1) and np.allclose(R @ R.T, np.eye(3))):
            raise TypeError('R is not a rotation matrix.')
        else:
            # rotate points
            self._points[:, :3] = np.einsum('ij,kj->ki', R, self._points[:, :3])
            # rotate lattice vectors
            self.latticeVectors = list(np.einsum('ij,kj->ki', R, self.latticeVectors))
            # update reciprocal vectors to reflect rotation
            self._makeReciprocalVectors()


class bravaisLattice(rhombicLattice):

    def __init__(self, latticeVectors, size, b=None):
        """
        Create a Bravais lattice. Lattice with one atom in the unit cell.

        See rhombicLattice for methods and attributes.

        Parameters
        ----------
        latticeVectors : list of array 1x3
            Lattice vectors defining the translation of the unit cell along its principal axes.
        size :3x integer, integer
            A list of integers describing the size in direction of the respective latticeVectors.
            Size is symmetric around zero in interval [-i,..,i] with length 2i+1.
            If one integer is given it is used for all 3 dimensions.
        b : list of float
            Corresponding scattering length of atoms in the unit cell.

        Returns
        -------
        lattice object
            .array  grid points as numpy array

        """
        rhombicLattice.__init__(self, latticeVectors, size, unitCellAtoms=[np.r_[0, 0, 0]], b=b)
        self._type = 'bravais'


class scLattice(bravaisLattice):

    def __init__(self, abc, size, b=None):
        """
        Simple Cubic lattice.

        See rhombicLattice for methods.

        Parameters
        ----------
        abc : float
            Lattice constant of unit cell.
        size : 3x integer, integer
            A list of integers describing the size in direction of the respective latticeVectors.
            Size is symmetric around zero in interval [-i,..,i] with length 2i+1.
            If one integer is given it is used for all 3 dimensions.
        b : list of float
            Corresponding scattering length of atoms in the unit cell.

        Returns
        -------
        lattice object
            .array  grid points as numpy array

        Examples
        --------
        ::

         import jscatter as js
         grid=js.sf.bccLattice(1.2,1)
         grid.show(2)

        """
        latticeVectors = [abc * np.r_[1., 0., 0.],
                          abc * np.r_[0., 1., 0.],
                          abc * np.r_[0., 0., 1.]]
        if isinstance(size, numbers.Number):
            size = [size] * 3
        bravaisLattice.__init__(self, latticeVectors, size, b=b)
        self._type = 'sc'


class bccLattice(rhombicLattice):

    def __init__(self, abc, size, b=None):
        """
        Body centered cubic lattice.

        See rhombicLattice for methods.

        Parameters
        ----------
        abc : float
            Lattice constant of unit cell.
        size : 3x integer, integer
            A list of integers describing the size in direction of the respective latticeVectors.
            Size is symmetric around zero in interval [-i,..,i] with length 2i+1.
            If one integer is given it is used for all 3 dimensions.
        b : list of float
            Corresponding scattering length of atoms in the unit cell.

        Returns
        -------
        lattice object
            .array  grid points as numpy array

        Examples
        --------
        ::

         import jscatter as js
         grid=js.sf.bccLattice(1.2,1)
         grid.show(2)

        """
        unitCellAtoms = [np.r_[0, 0, 0], np.r_[0.5, 0.5, 0.5]]
        latticeVectors = [abc * np.r_[1., 0., 0.],
                          abc * np.r_[0., 1., 0.],
                          abc * np.r_[0., 0., 1.]]
        if isinstance(size, numbers.Number):
            size = [size] * 3
        rhombicLattice.__init__(self, latticeVectors, size, unitCellAtoms, b=b)
        self._type = 'bcc'


class fccLattice(rhombicLattice):

    def __init__(self, abc, size, b=None):
        """
        Face centered cubic lattice.

        See rhombicLattice for methods.

        Parameters
        ----------
        abc : float
            Lattice constant of unit cell.
        size : 3x integer, integer
            A list of integers describing the size in direction of the respective latticeVectors.
            Size is symmetric around zero in interval [-i,..,i] with length 2i+1.
            If one integer is given it is used for all 3 dimensions.
        b : list of float
            Corresponding scattering length of atoms in the unit cell.

        Returns
        -------
        lattice object
            .array  grid points as numpy array

        Examples
        --------
        ::

         import jscatter as js
         grid=js.sf.fccLattice(1.2,1)
         grid.show(2)

        """
        unitCellAtoms = [np.r_[0, 0, 0],
                         np.r_[0, 0.5, 0.5],
                         np.r_[0.5, 0, 0.5],
                         np.r_[0.5, 0.5, 0]]
        latticeVectors = [abc * np.r_[1., 0., 0.],
                          abc * np.r_[0., 1., 0.],
                          abc * np.r_[0., 0., 1.]]
        if isinstance(size, numbers.Number):
            size = [size] * 3
        rhombicLattice.__init__(self, latticeVectors, size, unitCellAtoms, b=b)
        self._type = 'fcc'


class diamondLattice(rhombicLattice):

    def __init__(self, abc, size, b=None):
        """
        Diamond cubic lattice with 8 atoms in unit cell.

        See rhombicLattice for methods.

        Parameters
        ----------
        abc : float
            Lattice constant of unit cell.
        size : 3x integer, integer
            A list of integers describing the size in direction of the respective latticeVectors.
            Size is symmetric around zero in interval [-i,..,i] with length 2i+1.
            If one integer is given it is used for all 3 dimensions.
        b : list of float
            Corresponding scattering length of atoms in the unit cell.

        Returns
        -------
        lattice object
            .array  grid points as numpy array

        Examples
        --------
        ::

         import jscatter as js
         grid=js.sf.diamondLattice(1.2,1)
         grid.show(2)

        """
        unitCellAtoms = [np.r_[0, 0, 0],
                         np.r_[0.5, 0.5, 0],
                         np.r_[0, 0.5, 0.5],
                         np.r_[0.5, 0, 0.5],
                         np.r_[1 / 4., 1 / 4., 1 / 4.],
                         np.r_[3 / 4., 3 / 4., 1 / 4.],
                         np.r_[1 / 4., 3 / 4., 3 / 4.],
                         np.r_[3 / 4., 1 / 4., 3 / 4.]]
        latticeVectors = [abc * np.r_[1., 0., 0.],
                          abc * np.r_[0., 1., 0.],
                          abc * np.r_[0., 0., 1.]]
        if isinstance(size, numbers.Number):
            size = [size] * 3
        rhombicLattice.__init__(self, latticeVectors, size, unitCellAtoms, b=b)
        self._type = 'diamond'


class hexLattice(rhombicLattice):

    def __init__(self, ab, c, size, b=None):
        """
        Hexagonal lattice.

        See rhombicLattice for methods.

        Parameters
        ----------
        ab,c : float
            Lattice constant of unit cell.
            ab is distance in hexagonal plane, c perpendicular.
            For c/a = (8/3)**0.5 the hcp structure
        size : 3x integer, integer
            A list of integers describing the size in direction of the respective latticeVectors.
            Size is symmetric around zero in interval [-i,..,i] with length 2i+1.
            If one integer is given it is used for all 3 dimensions.
        b : list of float
            Corresponding scattering length of atoms in the unit cell.

        Returns
        -------
        lattice object
            .array  grid points as numpy array

        Examples
        --------
        ::

         import jscatter as js
         grid=js.sf.hexLattice(1.,2,[2,2,2])
         grid.show(2)

        """

        latticeVectors = [np.r_[ab, 0., 0.],
                          np.r_[0.5 * ab, 3 ** 0.5 / 2 * ab, 0.],
                          np.r_[0., 0., c]]
        unitCellAtoms = [np.r_[0, 0, 0]]
        if isinstance(size, numbers.Number):
            size = [size] * 3
        rhombicLattice.__init__(self, latticeVectors, size, unitCellAtoms, b=b)
        self._type = 'hex'


class honeycombLattice(rhombicLattice):

    def __init__(self, ab, c, size, b=None):
        """
        Honeycomb lattice e.g for graphene.

        See rhombicLattice for methods.

        Parameters
        ----------
        ab,c : float
            Lattice constants of unit cell.
            ab is distance between nearest neighbors in honeycomb plane, c perpendicular.
        size : 3x integer, integer
            A list of integers describing the size in direction of the respective latticeVectors.
            Size is symmetric around zero in interval [-i,..,i] with length 2i+1.
            If one integer is given it is used for all 3 dimensions.
        b : list of float
            Corresponding scattering length of atoms in the unit cell.

        Returns
        -------
        lattice object
            .array  grid points as numpy array

        Notes
        -----
        Similar peak positions as hexagonalLattice with ab*3**0.5
        with same peak height for [1,2,n] and smaller peak height in [2,2,n].

        Examples
        --------
        ::

         import jscatter as js
         grid=js.sf.honeycombLattice(1.,2,[2,2,2],[1,2])
         grid.show(5)

        """
        latticeVectors = [ab/2 * np.r_[3, -3 ** 0.5, 0.],
                          ab/2 * np.r_[3,  3 ** 0.5, 0.],
                          np.r_[0., 0., c]]
        unitCellAtoms = [np.r_[0, 0, 0], np.r_[1, 1, 0] / 3]
        if isinstance(size, numbers.Number):
            size = [size] * 3
        rhombicLattice.__init__(self, latticeVectors, size, unitCellAtoms, b=b)
        self._type = 'hex'


class hcpLattice(rhombicLattice):

    def __init__(self, ab, size, b=None):
        """
        Hexagonal closed packed lattice.

        See rhombicLattice for methods.

        Parameters
        ----------
        ab : float
            Lattice constant of unit cell.
            ab is distance in hexagonal plane, c = ab* (8/3)**0.5
        size : 3x integer, integer
            A list of integers describing the size in direction of the respective latticeVectors.
            Size is symmetric around zero in interval [-i,..,i] with length 2i+1.
            If one integer is given it is used for all 3 dimensions.
        b : list of float
            Corresponding scattering length of atoms in the unit cell.

        Returns
        -------
        lattice object
            .array  grid points as numpy array

        Examples
        --------
        ::

         import jscatter as js
         grid=js.sf.hcpLattice(1.2,[3,3,1])
         grid.show(2)

        """
        c = ab * (8 / 3.) ** 0.5
        latticeVectors = [np.r_[ab, 0., 0.],
                          np.r_[0.5 * ab, 3 ** 0.5 / 2 * ab, 0.],
                          np.r_[0., 0., c]]
        unitCellAtoms = [np.r_[0, 0, 0],
                         np.r_[1 / 3., 1 / 3., 0.5]]

        if isinstance(size, numbers.Number):
            size = [size] * 3
        rhombicLattice.__init__(self, latticeVectors, size, unitCellAtoms, b=b)
        self._type = 'hcp'


class sqLattice(bravaisLattice):

    def __init__(self, ab, size, b=None):
        """
        Simple 2D square lattice.

        See rhombicLattice for methods.

        Parameters
        ----------
        ab : float
            Lattice constant of unit cell.
        size : 2x integer, integer
            A list of integers describing the size in direction of the respective latticeVectors.
            Size is symmetric around zero in interval [-i,..,i] with length 2i+1.
            If one integer is given it is used for all 3 dimensions.
        b : list of float
            Corresponding scattering length of atoms in the unit cell.

        Returns
        -------
        lattice object
            .array  grid points as numpy array

        Examples
        --------
        ::

         import jscatter as js
         grid=js.sf.sqLattice(1.2,1)
         grid.show(2)

        """
        latticeVectors = [ab * np.r_[1., 0., 0.],
                          ab * np.r_[0., 1., 0.]]
        unitCellAtoms = [np.r_[0, 0]]  # only 2D
        if isinstance(size, numbers.Number):
            size = [size] * 2
        rhombicLattice.__init__(self, latticeVectors, size, unitCellAtoms, b=b)
        self._type = 'sq'


class hex2DLattice(bravaisLattice):

    def __init__(self, ab, size, b=None):
        """
        Simple 2D hexagonal lattice.

        See rhombicLattice for methods.

        Parameters
        ----------
        ab : float
            Lattice constant of unit cell.
        size : 2x integer, integer
            A list of integers describing the size in direction of the respective latticeVectors.
            Size is symmetric around zero in interval [-i,..,i] with length 2i+1.
            If one integer is given it is used for all 3 dimensions.
        b : list of float
            Corresponding scattering length of atoms in the unit cell.

        Returns
        -------
        lattice object
            .array  grid points as numpy array

        Examples
        --------
        ::

         import jscatter as js
         grid=js.sf.hex2DLattice(1.2,1)
         grid.show(2)

        """
        latticeVectors = [np.r_[ab, 0., 0.],
                          np.r_[-0.5 * ab, 3 ** 0.5 / 2 * ab, 0.]]
        unitCellAtoms = [np.r_[0, 0]]  # only 2D
        if isinstance(size, numbers.Number):
            size = [size] * 2
        rhombicLattice.__init__(self, latticeVectors, size[:2], unitCellAtoms, b=b)
        self._type = 'hex'


class lamLattice(bravaisLattice):

    def __init__(self, a, size, b=None):
        """
        1D lamellar lattice.

        See rhombicLattice for methods.

        Parameters
        ----------
        a : float
            Lattice constant of unit cell.
        size : 1x integer, integer
            A list of integers describing the size in direction of the respective latticeVectors.
            Size is symmetric around zero in interval [-i,..,i] with length 2i+1.
            If one integer is given it is used for all 3 dimensions.
        b : list of float
            Corresponding scattering length of atoms in the unit cell.

        Returns
        -------
        lattice object
            .array  grid points as numpy array

        Examples
        --------
        ::

         import jscatter as js
         grid=js.sf.lamLattice(1.2,1)
         grid.show(2)

        """
        latticeVectors = [np.r_[a, 0., 0.]]
        unitCellAtoms = [np.r_[0]]  # only 1D
        if isinstance(size, numbers.Number):
            size = [size] * 1
        rhombicLattice.__init__(self, latticeVectors, size[:1], unitCellAtoms, b=b)
        self._type = 'lam'


class latticeFromCIF(rhombicLattice):
    """ see init"""

    def __init__(self, structure, size=[1, 1, 1], typ='x'):
        r"""
        Create lattice as defined in CIF (Crystallographic Information Format) file or
        a pymatgen structure.

        Parameters
        ----------
        structure : str, pymatgen.structure
            Filename of the CIF file or structure read by pymatgen as
            :code:`structure = pymatgen.core.Structure.from_file(js.examples.datapath+'/1011053.cif')`
        typ : 'xray','neutron'
            scattering length for coherent xray or neutron scattering.
        size : 3x int
            Size of the lattice in direction of lattice vectors

        Notes
        -----
        - Pymatgen (Python Materials Genomics) is a robust, open-source Python library for
          materials analysis [1]_. See `Pymatgen <https://pymatgen.org/index.html#>`_
          ::

           # Simply install by
           pip install pymatgen

        - CIF files can be found in the
          `Crystallography Open Database <http://www.crystallography.net/cod/>`_

        - Pymatgen allows site occupancy which is taken here into account as average scattering length per site.

        - Pymatgen allows reading of CIF files or to get directly a structure from
          `Materials Project <https://materialsproject.org/>`_
          or `Crystallography Open Database <http://www.crystallography.net/cod/>`_ .

          Look at
          `Examples <http://matgenb.materialsvirtuallab.org/2013/01/01/Getting-crystal-structures-from-online-sources.html>`_
          ::

           from pymatgen.ext.cod import COD
           cod = COD()
           # SiC silicon carbide (carborundum or Moissanite)
           sic = cod.get_structure_by_id(1011053)
           sicc=js.sf.latticeFromCIF(sic)
           # silicon
           si = cod.get_structure_by_id(9008566)

        - If Bragg peaks or reciprocal lattice points are missing try to increase the size parameter
          as these points might belong to higher order lattice planes.
          E.g. for SiC (silicon carbide) the second peak at :math:`2\theta=35.6^{\circ}` belongs to hkl = 006 or
          the  8th order peak at :math:`2\theta=60.1^{\circ}` belongs to hkl = 108 .


        Examples
        --------
        The example is for silicon carbide. Please compare to `Moissanite
        <http://rruff.info/repository/sample_child_record_powder/by_minerals/Moissanite__R061083-9__Powder__DIF_File__9401.txt>`_
        ::

         import jscatter as js
         import pymatgen
         siliconcarbide = pymatgen.core.Structure.from_file(js.examples.datapath+'/1011053.cif')
         sic=js.sf.latticeFromCIF(siliconcarbide)
         sic.getScatteringAngle(wavelength=0.13141,size=13)

         from pymatgen.ext.cod import COD
         cod = COD()
         # 3C-SiC silicon carbide (carborundum or Moissanite)
         siliconcarbide = cod.get_structure_by_id(1011031)
         sic=js.sf.latticeFromCIF(siliconcarbide)
         sic.getScatteringAngle(size=13)


        References
        ----------
        .. [1] Python Materials Genomics (pymatgen) : A Robust, Open-Source Python Library for Materials Analysis.
               Ong et al
               Computational Materials Science, 2013, 68, 314–319.
               doi:10.1016/j.commatsci.2012.10.028

        """
        #
        if isinstance(structure, str):
            if pymatgenfound:
                cif = pymatgen.core.Structure.from_file(structure)
            else:
                raise Exception('Please install pymatgen to read CIF files. See doc of latticeFromCIF')
        else:
            # assume it is a pymatgen structure
            cif = structure

        if typ[0] == 'n':
            # coherent neutron scattering length using fractional occupancies
            b=[]
            for specOcc in cif.species_and_occu:
                names = [e.name for e in specOcc.elements]
                b_frac = [formel.Nscatlength[n.lower()][0]*specOcc.element_composition.get_atomic_fraction(n)
                          for n in names]
                b.append(np.sum(b_frac))
        else:
            # xray scattering length using number of electrons
            # species_and_occu averages occupancy of different atoms on sites
            b = [specOcc.total_electrons * formel.felectron for specOcc in cif.species_and_occu]

        rhombicLattice.__init__(self, latticeVectors=cif.lattice.matrix * 0.1, size=size,
                                unitCellAtoms=cif.frac_coords, b=b)
        self.cif = cif


class latticeFromMDA(rhombicLattice):
    """ see init"""

    def __init__(self, atoms, latticeVectors=None, size=[1, 1, 1], typ='x'):
        r"""
        Create lattice with a unit cell defined from MDAnalysis universe atomgroup.

        Parameters
        ----------
        atoms : MDAnalysis universe atomgroup
            Atomsgroup of universe created with MDAnalysis or bio.scatteringUniverse
        latticeVectors : 3x array(3)
            Lattice vectors spanning the unit cell enclosing the atomgroup.
        typ : 'xray','neutron'
            scattering length for coherent xray or neutron scattering.
        size : 3x int
            Size of the lattice in direction of lattice vectors


        Examples
        --------
        We look at an alpha helix ordered in a hexagonal lattice.
        We need first to orient the helix to fit into the lattice with defined orientation.
        The aim is to calculate the structure factor.
        ::

         import jscatter as js
         import numpy as np
         from scipy.spatial.transform import Rotation as Rot
         from scipy import linalg as la

         uniR=js.bio.scatteringUniverse(js.examples.datapath+'/arg61.pdb',addHydrogen=False)

         # rotate the helix parallel to the Z axis
         uniR.atoms.positions = uniR.atoms.positions-uniR.atoms.center_of_geometry()
         up = uniR.select_atoms('name CA and resnum 2-8')
         vup = up.positions.mean(axis=0)
         rotv = np.cross(vup,[0,0,1])
         rotv = rotv/la.norm(rotv)
         r = Rot.from_rotvec( -np.arctan(la.norm(vup[:2])/vup[2]) * rotv)
         uniR.atoms.positions =  uniR.atoms.positions @ r.as_matrix()
         uniR.atoms.positions = uniR.atoms.positions - np.r_[0,0,uniR.atoms.positions[:,2].min()]
         # uniR.view(viewer='pymol')

         # define lattice constant
         hex_c = 13  # bit larger then helix
         hex_a = 2   # bit larger than radius

         # get lattice vectors
         v1,v2,v3 = js.sf.hexLattice(hex_a,hex_c,1).latticeVectors

         # create lattice with universe atoms in unit cell
         PLA = js.sf.latticeFromMDA(uniR.atoms,[v1,v2,v3],size=[1,1, 0])
         # PLA.show()

         # now we use latticeStructureFactor to calculate the lattice structure factor
         p = js.grace()
         p.title('poly arginine alpha helix in hexagonal lattice')
         p.subtitle('hexagonal peak intensity modified by unit cell atoms')

         q = np.r_[0.1:25:1000j]
         sf = js.sf.latticeStructureFactor(q, lattice=PLA,domainsize=60, rmsd=0.02, lg=1, hklmax=6,wavelength=0.15406)
         p.plot(sf.X,sf.Y + 0.03,sy=[1,0.3,1],li=1)

         p.xaxis(label='q / nm\S-1',scale='log',min=0.1,max=20)
         p.yaxis(label='I(q)',scale='log',min=0.02,max=5)
         p.text(string='lamellar peaks',x=0.7,y=1)
         p.text(string='hexagonal peaks',x=3.6,y=0.4)
         # p.save(js.examples.imagepath+'/PLALatticeSf.jpg',size=[600,600])

        .. image:: ../../examples/images/PLALatticeSf.jpg
         :width: 50 %
         :align: center
         :alt: rmsa


        """
        if typ[0] == 'n':
            b = atoms.fan0()
        else:
            b = atoms.fax0()

        frac_coordinates = atoms.posnm @ la.inv(latticeVectors)
        rhombicLattice.__init__(self, latticeVectors=latticeVectors, size=size,
                                unitCellAtoms=frac_coordinates, b=b)




import jscatter as js
import numpy as np

# Here we examine the effect of instrumental smearing for SAX (Kratky Camera, line! ) and SANS
# and how we can use the Lake algorithm for desmearing.

# some data
q = np.r_[0.01:7:0.01]
# obj = js.ff.sphere(q,5)
data = js.ff.ellipsoid(q, 2, 3)
# add background
data.Y += 2
# load data for beam width profile
empty = js.dA(js.examples.datapath + '/buffer_averaged_corrected_despiked.pdh', usecols=[0, 1],
              lines2parameter=[2, 3, 4])
# read beam length profile measurement for a slit (Kratky Camera)
beam = js.dA(js.examples.datapath + '/BeamProfile.pdh', usecols=[0, 1], lines2parameter=[2, 3, 4])

# fit beam width for line collimation with semitransparent beam stop
bwidth = js.sas.getBeamWidth(empty, 'auto')

# prepare measured beamprofile from beam measurement
mbeam = js.sas.prepareBeamProfile(beam, a=2, b=1, bxw=bwidth, dIW=1.)
# prepare profile with trapezoidal shape using explicit given parameters (using parameters from above to have equal)
tbeam = js.sas.prepareBeamProfile('trapez', a=mbeam.a, b=mbeam.b, bxw=bwidth, dIW=1)
# prepare profile SANS a la Pedersen in point collimation. This can be used for SAXS with smaller apertures too!
Sbeam = js.sas.prepareBeamProfile('SANS', detDist=2000, wavelength=0.4, wavespread=0.15)

if 0:
    p = js.sas.plotBeamProfile(mbeam)
    p = js.sas.plotBeamProfile(mbeam, p)

# smear
datasm = js.sas.smear(unsmeared=data, beamProfile=mbeam)
datast = js.sas.smear(unsmeared=data, beamProfile=tbeam)
datasS = js.sas.smear(unsmeared=data, beamProfile=Sbeam)
# add noise
datasm.Y += np.random.normal(0, 0.5, len(datasm.X))
datast.Y += np.random.normal(0, 0.5, len(datast.X))
datasS.Y += np.random.normal(0, 0.5, len(datasS.X))

# desmear again
ws = 11
NI = -15
dsm = js.sas.desmear(datasm, mbeam, NIterations=NI, windowsize=ws)
dst = js.sas.desmear(datast, tbeam, NIterations=NI, windowsize=ws)
dsS = js.sas.desmear(datasS, Sbeam, NIterations=NI, windowsize=ws)

# compare
p = js.grace(2, 1.4)
p.plot(data, sy=[1, 0.3, 3], le='original ellipsoid')
p.plot(dst, sy=0, li=[1, 2, 1], le='desmeared SAX line collimation')
p.plot(dsS, sy=0, li=[1, 2, 2], le='desmeared SANS')
p.plot(datasm, li=[3, 2, 1], sy=0, le='smeared SAX line collimation')
p.plot(datasS, li=[3, 2, 4], sy=0, le='smeared SANS')
p.yaxis(max=1e4, min=0.1, scale='l', label='Intensity / a.u.', size=1.7)
p.xaxis(label=r'q / nm\S-1', size=1.7)
p.legend(x=3, y=5500, charsize=1.7)
p.title('Smeared ellipsoid and desmearing by Lake algorithm')

# The conclusion is to better fit smeared models than to desmear and fit unsmeared models.
p.save('SASdesmearing.png')



# -*- coding: utf-8 -*-
# written by Ralf Biehl at the Forschungszentrum Jülich ,
# Jülich Center for Neutron Science 1 and Institute of Complex Systems 1
#    Jscatter is a program to read, analyse and plot data
#    Copyright (C) 2015-2024  Ralf Biehl
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Physical equations and useful formulas as quadrature of vector functions, parallel execution,
viscosity, compressibility of water, scatteringLengthDensityCalc or sedimentationProfile.
Use scipy.constants for physical constants.

- Each topic is not enough for a single module, so this is a collection.
- All scipy functions can be used. See http://docs.scipy.org/doc/scipy/reference/special.html.
- Statistical functions http://docs.scipy.org/doc/scipy/reference/stats.html.

Mass and scattering length of all elements in Elements are taken from :
 - Mass: http://www.chem.qmul.ac.uk/iupac/AtWt/
 - Neutron scattering length: http://www.ncnr.nist.gov/resources/n-lengths/list.html

Units converted to amu for mass and nm for scattering length.

"""

import functools
import os
import pickle
from collections import deque

import numpy as np
from numpy import linalg as la
import scipy
import scipy.integrate
import scipy.signal
import scipy.optimize

from .. import parallel


_path_ = os.path.realpath(os.path.dirname(__file__))

__all__ = ['memoize', 'rotationMatrix', 'xyz2rphitheta', 'rphitheta2xyz', 'qEwaldSphere',
           'loglist', 'smooth']

#: Variable to allow printout for debugging as if debug:print('message')
debug = False


# noinspection PyIncorrectDocstring
def memoize(**memkwargs):
    """
    A least-recently-used cache decorator to cache expensive function evaluations.

    Memoize caches results and retrieves from cache if same parameters are used again.
    This can speedup computation in a model if a part is computed with same parameters several times.
    During fits it may be faster to calc result for a list and take from cache.
    See also https://docs.python.org/3/library/functools.html  cache or lru_cache

    Parameters
    ----------
    function : function
        Function to evaluate as e.g. f(Q,a,b,c,d)
    memkwargs : dict
        Keyword args with substitute values to cache for later interpolation. Empty for normal caching of a function.
        E.g. memkwargs={'Q':np.r_[0:10:0.1],'t':np.r_[0:100:5]} caches with these values.
        The needed values can be interpolated from the returned result. See example below.
    maxsize : int, default 128
        maximum size of the cache. Last is dropped.

    Returns
    -------
    function
        cached function with new methods
         - last(i) to retrieve the ith evaluation result in cache (last is i=-1).
         - clear() to clear the cached results.
         - hitsmisses counts hits and misses.

    Notes
    -----
    Only keyword arguments for the memoized function are supported!!!!
    Only one attribute and X are supported for fitting as .interpolate works only for two cached attributes.



    Examples
    --------
    The example uses a model that computes like I(q,n,..)=F(q)*B(t,n,..).
    F(q) is cheap to calculate B(t,n,..) not. In the following its better to calc
    the function for a list of q , put it to cache and take in the fit from there.
    B is only calculated once inside of the function.

    Use it like this::

     import jscatter as js
     import numpy as np

     # define some data
     TT=js.loglist(0.01,80,30)
     QQ=np.r_[0.1:1.5:0.15]
     # in the data we have 'q' and 'X'
     data=js.dynamic.finiteZimm(t=TT,q=QQ,NN=124,pmax=100,tintern=10,l=0.38,Dcm=0.01,mu=0.5,viscosity=1.001,Temp=300)

     # makes a unique list of all X values    -> interpolation is exact for X
     # one may also use a smaller list of values and only interpolate
     tt=list(set(data.X.flatten));tt.sort()

     # define memoized function which will always use the here defined q and t
     # use correct values from data for q     -> interpolation is exact for q
     memfZ=js.formel.memoize(q=data.q,t=tt)(js.dynamic.finiteZimm)

     def fitfunc(Q,Ti,NN,tint,ll,D,mu,viscosity,Temp):
        # use the memoized function as usual (even if given t and q are used from above definition)
        res= memfZ(NN=NN,tintern=tint,l=ll,Dcm=D,pmax=40,mu=mu,viscosity=viscosity,Temp=Temp)
        # interpolate to the here needed q and t (which is X)
        resint=res.interpolate(q=Q,X=Ti,deg=2)[0]
        return resint

     # do the fit
     data.setlimit(tint=[0.5,40],D=[0,1])
     data.makeErrPlot(yscale='l')
     NN=20
     data.fit(model=fitfunc,
              freepar={'tint':10,'D':0.1,},
              fixpar={'NN':20,'ll':0.38/(NN/124.)**0.5,'mu':0.5,'viscosity':0.001,'Temp':300},
              mapNames={'Ti':'X','Q':'q'},)

    Second example

    Use memoize as a decorator (@ in front) acting on the following function.
    This is a shortcut for the above and works in the same way
    ::

     # define the function to memoize
     @js.formel.memoize(Q=np.r_[0:3:0.2],Time=np.r_[0:50:0.5,50:100:5])
     def fZ(Q,Time,NN,tintern,ll,Dcm,mu,viscosity,Temp):
         # finiteZimm accepts t and q as array and returns a dataList with different Q and same X=t
         res=js.dynamic.finiteZimm(t=Time,q=Q,NN=NN,pmax=20,tintern=tintern,
                               l=ll,Dcm=Dcm,mu=mu,viscosity=viscosity,Temp=Temp)
         return res

     # define the fitfunc
     def fitfunc(Q,Ti,NN,tint,ll,D,mu,viscosity,Temp):
        #this is the cached result for the list of Q
        res= fZ(Time=Ti,Q=Q,NN=NN,tintern=tint,ll=ll,Dcm=D,mu=mu,viscosity=viscosity,Temp=Temp)
        # interpolate for the single Q value the cached result has again 'q'
        return res.interpolate(q=Q,X=Ti,deg=2)[0]

     # do the fit
     data.setlimit(tint=[0.5,40],D=[0,1])
     data.makeErrPlot(yscale='l')
     data.fit(model=fitfunc,
              freepar={'tint':6,'D':0.1,},
              fixpar={'NN':20,'ll':0.38/(20/124.)**0.5,'mu':0.5,'viscosity':0.001,'Temp':300},
              mapNames={'Ti':'X','Q':'q'})
     # the result depends on the interpolation;


    """
    cachesize = memkwargs.pop('maxsize', 128)

    def _memoize(function):
        function.hitsmisses = [0, 0]
        cache = function.cache = {}
        deck = function.deck = deque([], maxlen=cachesize)
        function.last = lambda i=-1: function.cache[function.deck[i]]

        def clear():
            while len(function.deck) > 0:
                del function.cache[function.deck.pop()]
            function.hitsmisses = [0, 0]

        function.clear = clear

        @functools.wraps(function)
        def _memoizer(*args, **kwargs):
            # make new
            nkwargs = dict(kwargs, **memkwargs)
            key = pickle.dumps(args) + pickle.dumps(nkwargs)
            if key in cache:
                function.hitsmisses[0] += 1
                deck.remove(key)
                deck.append(key)
                return cache[key]
            else:
                function.hitsmisses[1] += 1
                cache[key] = function(*args, **nkwargs)
                if len(deck) >= cachesize:
                    del cache[deck.popleft()]
                deck.append(key)
                return cache[key]

        return _memoizer

    return _memoize


def rotationMatrix(vector, angle):
    """
    Create a rotation matrix corresponding to rotation around vector v by a specified angle.

    .. math::  R = vv^T + cos(a) (I - vv^T) + sin(a) skew(v)
    See Notes for scipy rotation matrix.

    Parameters
    ----------
    vector : array
        Rotation around a general  vector
    angle : float
        Angle in rad

    Returns
    -------
    array
        Rotation matrix

    Notes
    -----
    A convenient way to define more complex rotations is found in
    `scipy.spatial.transform.Rotation
    <https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.transform.Rotation.html>`_ .
    E.g. by Euler angles and returned as rotation matrix ::

     from scipy.spatial.transform import Rotation as Rot
     R=Rot.from_euler('YZ',[90,10],1).as_matrix()

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle

    Examples
    --------
    Examples show how to use a  rotation matrix with vectors.
    ::

     import jscatter as js
     import numpy as np
     from matplotlib import pyplot
     R=js.formel.rotationMatrix([0,0,1],np.deg2rad(-90))
     v=[1,0,0]
     # rotated vector
     rv=np.dot(R,v)
     #
     # rotate Fibonacci Grid
     qfib=js.formel.fibonacciLatticePointsOnSphere(300,1)
     qfib=qfib[qfib[:,2]<np.pi/2,:]                       # select half sphere
     qfib[:,2]*=(30/90.)                                  # shrink to cone of 30°
     qfx=js.formel.rphitheta2xyz(qfib)                    # transform to cartesian
     v = [0,1,0]                                          # rotation vector
     R=js.formel.rotationMatrix(v,np.deg2rad(90))        # rotation matrix around v axis
     Rfx=np.einsum('ij,kj->ki',R,qfx)                     # do rotation
     fig = pyplot.figure()
     ax = fig.add_subplot(111, projection='3d')
     sc=ax.scatter(qfx[:,0], qfx[:,1], qfx[:,2], s=2, color='r')
     sc=ax.scatter(Rfx[:,0], Rfx[:,1], Rfx[:,2], s=2, color='b')
     ax.scatter(0,0,0, s=55, color='g',alpha=0.5)
     ax.quiver([0],[0],[0],*v,color=['g'])
     fig.axes[0].set_title('rotate red points to blue around vector (green)')
     pyplot.show(block=False)
     # fig.savefig(js.examples.imagepath+'/rotationMatrix.jpg')

    .. image:: ../../examples/images/rotationMatrix.jpg
     :align: center
     :height: 300px
     :alt: sq2gr



    """
    normd = np.linalg.norm(vector)
    assert normd != 0, 'vector norm should be >0.'

    d = np.array(vector, dtype=np.float64)
    d = d / normd
    eye = np.eye(3, dtype=np.float64)
    ddt = np.outer(d, d)
    skew = np.array([[0, -d[2], d[1]],
                     [d[2], 0, -d[0]],
                     [-d[1], d[0], 0]], dtype=np.float64)
    mtx = np.cos(angle) * eye + np.sin(angle) * skew + (1 -np.cos(angle)) * ddt
    return mtx


def xyz2rphitheta(XYZ, transpose=False):
    """
    Transformation cartesian coordinates [X,Y,Z] to spherical coordinates [r,phi,theta].

    Parameters
    ----------
    XYZ : array Nx3
        Coordinates with [x,y,z]  ( XYZ.shape[1]==3).
    transpose : bool
        Transpose XYZ before transformation.

    Returns
    -------
    array Nx3
        Coordinates with [r,phi,theta]
         - phi   : float   azimuth     -pi < phi < pi
         - theta : float   polar angle  0 < theta  < pi
         - r     : float   length

    Examples
    --------
    Single coordinates
    ::

     js.formel.xyz2rphitheta([1,0,0])

    Transform Fibonacci lattice on sphere to xyz coordinates
    ::

     rpc=js.formel.randomPointsInCube(10)
     js.formel.xyz2rphitheta(rpc)

    Tranformation 2D X,Y plane coordinates to r,phi coordinates (Z=0)
    ::

     rp=js.formel.xyz2rphitheta([data.X,data.Z,abs(data.X*0)],transpose=True) )[:,:2]

    """
    xyz = np.array(XYZ, ndmin=2)
    if transpose:
        xyz = xyz.T
    assert xyz.shape[1] == 3, 'XYZ second dimension should be 3. Transpose it?'
    rpt = np.empty(xyz.shape)
    rpt[:, 0] = la.norm(xyz, axis=1)
    rpt[:, 1] = np.arctan2(xyz[:, 1], xyz[:, 0])  # arctan2 is special function for this purpose
    rpt[:, 2] = np.arctan2(la.norm(xyz[:, :-1], axis=1), xyz[:, 2])
    return np.array(rpt.squeeze(), ndmin=np.ndim(XYZ))


def rphitheta2xyz(RPT, transpose=False):
    """
    Transformation  spherical coordinates [r,phi,theta]  to cartesian coordinates [x,y,z].

    Parameters
    ----------
    RPT : array Nx3
        Coordinates with [r,phi,theta]
         - r     : float   length
         - phi   : float   azimuth     -pi < phi < pi
         - theta : float   polar angle  0 < theta  < pi
    transpose : bool
        Transpose RPT before transformation.

    Returns
    -------
    array Nx3
        [x,y,z] coordinates

    """
    rpt = np.array(RPT, ndmin=2)
    if transpose:
        rpt = rpt.T
    assert rpt.shape[1] == 3, 'RPT second dimension should be 3. Transpose it?'
    xyz = np.zeros(rpt.shape)
    xyz[:, 0] = rpt[:, 0] * np.cos(rpt[:, 1]) * np.sin(rpt[:, 2])
    xyz[:, 1] = rpt[:, 0] * np.sin(rpt[:, 1]) * np.sin(rpt[:, 2])
    xyz[:, 2] = rpt[:, 0] * np.cos(rpt[:, 2])
    return np.array(xyz.squeeze(), ndmin=np.ndim(RPT))


def qEwaldSphere(q, wavelength=0.15406, typ=None, N=60):
    r"""
    Points on Ewald sphere with different distributions.

    :math:`q = \vec{k_s} -\vec{k_i} =4\pi/\lambda sin(\theta/2)` with
    :math:`\vec{k_i} =[0,0,1]` and :math:`|\vec{k_i}| =2\pi/\lambda`

    Use rotation matrix to rotate to specific orientations.

    Parameters
    ----------
    q : array,list
        Wavevectors units 1/nm
    wavelength : float
        Wavelength of radiation, default X-ray K_a.
    N : integer
        Number of points in intervals.
    typ : 'cart','ring','random' default='ring'
        Typ of q value distribution on Ewald sphere.
         - cart : Cartesian grid between -q_max,q_max with N points (odd to include zero).
         - ring : Given q values with N-points on rings of equal q.
         - random : N² random points on Ewald sphere between q_min and q_max.

    Returns
    -------
        array : 3xN [x,y,z] coordinates


    Examples
    --------
    ::

     import jscatter as js
     import numpy as np
     fig = js.mpl.figure(figsize=[8, 2.7],dpi=200)
     ax1 = fig.add_subplot(1, 4, 1, projection='3d')
     ax2 = fig.add_subplot(1, 4, 2, projection='3d')
     ax3 = fig.add_subplot(1, 4, 3, projection='3d')
     ax4 = fig.add_subplot(1, 4, 4, projection='3d')
     q0 = 2 * np.pi / 0.15406  # Ewald sphere radius |kin|

     q = np.r_[0.:2*q0:10j]
     qe = js.formel.qEwaldSphere(q)
     js.mpl.scatter3d(qe[0],qe[1],qe[2],ax=ax1,pointsize=1)
     ax1.set_title('equidistant q')

     q = 2*q0*np.sin(np.r_[0.:np.pi:10j]/2)
     qe = js.formel.qEwaldSphere(q)
     js.mpl.scatter3d(qe[0],qe[1],qe[2],ax=ax2,pointsize=1)
     ax2.set_title('equidistant angle')

     qe = js.formel.qEwaldSphere(q=[10],N=20,typ='cart')
     js.mpl.scatter3d(qe[0],qe[1],qe[2],ax=ax3,pointsize=1)
     ax3.set_title('cartesian grid')

     qe = js.formel.qEwaldSphere(q=[10,0.5*q0],N=60,typ='random')
     fig = js.mpl.scatter3d(qe[0],qe[1],qe[2],ax=ax4,pointsize=1)
     ax4.set_title('random min,max')
     #fig.savefig(js.examples.imagepath+'/qEwaldSphere.jpg')

    .. image:: ../../examples/images/qEwaldSphere.jpg
     :width: 90 %
     :align: center
     :alt: qEwaldSphere

    """
    q = np.array(q)  # scattering vector 4

    # Ewald Sphere radius
    q0 = 2 * np.pi / wavelength

    if typ == 'cart':
        # cartesian grid in x.y
        x = y = np.r_[-q.max():q.max():1j * N]
        xx, yy = np.meshgrid(x, y, indexing='ij')
        qx = xx.flatten()
        qy = yy.flatten()
        qz = (q0 ** 2 - qx ** 2 - qy ** 2) ** 0.5 - q0
        if np.all(np.isfinite(qz)):
            return np.stack([qx.flatten(), qy.flatten(), qz.flatten()])
        else:
            raise ValueError('q.max() range to large for this wavelength.')

    elif typ == 'random':
        mi = 2 * np.arcsin(q.min() / 2/ q0)
        ma = 2 * np.arcsin(q.max() / 2/ q0)
        ringarea = (1 - np.cos(ma))/2 - (1 - np.cos(mi))/2
        # increase number by 1/ringarea
        rps = rphitheta2xyz(parallel.randomPointsOnSphere(int(N ** 2 / ringarea), q0)).T - np.r_[0, 0, q0][:, None]
        qrps = la.norm(rps, axis=0)
        return rps[:, (q.min() < qrps) & (qrps < q.max())]

    else:
        # q to angle
        theta = 2 * np.arcsin(q / 2/ q0)
        phi = np.r_[0:np.pi * 2:1j * (N+1)][:-1]
        # q = ks - ki
        # assume ki=[0,0,1], theta is scattering angle, phi azimuth
        qx = q0 * np.sin(theta) * np.cos(phi)[:, None]
        qy = q0 * np.sin(theta) * np.sin(phi)[:, None]
        qz = q0 * (np.cos(theta) - np.ones_like(phi)[:, None])

        return np.stack([qx.flatten(), qy.flatten(), qz.flatten()])


def loglist(mini=0.1, maxi=5, number=100):
    """
    Log like sequence between mini and maxi.

    Parameters
    ----------
    mini,maxi : float, default 0.1, 5
        Start and endpoint.
    number : int, default 100
        Number of points in sequence.

    Returns
    -------
    ndarray

    """
    ll = np.r_[np.log((mini if mini != 0. else 1e-6)):
               np.log((maxi if maxi != 0 else 1.)):
               (number if number != 0 else 10) * 1j]

    return np.exp(ll)


def smooth(data, windowlen=7, window='flat'):
    """
    Smooth data by convolution with window function or fft/ifft.

    Smoothing based on position ignoring information on .X.

    Parameters
    ----------
    data : array, dataArray
        Data to smooth.
        If is dataArray the .Y is smoothed and returned.
    windowlen : int, default = 7
        The length/size of the smoothing window; should be an odd integer.
        Smaller 3 returns unchanged data.
        For 'fourier' the high frequency cutoff is 2*size_data/windowlen.
    window :  'hann', 'hamming', 'bartlett', 'blackman','gaussian','fourier','flattop' default ='flat'
        Type of window/smoothing.
         - 'flat' will produce a moving average smoothing.
         - 'gaussian' normalized Gaussian window with sigma=windowlen/7.
         - 'fourier' cuts high frequencies above cutoff frequency between rfft and irfft.


    Returns
    -------
    array (only the smoothed array)

    Notes
    -----
    'hann', 'hamming', 'bartlett', 'blackman','gaussian', 'flat' :
     These methods convolve a scaled window function with the signal.
     The signal is prepared by introducing reflected copies of the signal (with the window size)
     at both ends so that transient parts are minimized in the beginning and end part of the output signal.
     Adapted from SciPy/Cookbook.

     See
     https://docs.scipy.org/doc/scipy-1.11.0/reference/generated/scipy.signal.get_window.html#scipy.signal.get_window

    fourier :
     The real valued signal is mirrored at left side, Fourier transformed, the high frequencies are cut
     and the signal is back transformed.
     This is the simplest form as a hard cutoff frequency is used (ideal low pass filter)
     and may be improved using a specific window function in frequency domain.
     ::

      rft = np.fft.rfft(np.r_[data[::-1],data])
      rft[int(2*len(data)/windowlen):] = 0
      smoothed = np.fft.irfft(rft)

    Examples
    --------
    Usage:
    ::

     import jscatter as js
     import numpy as np
     t=np.r_[-5:5:0.01]
     data=np.sin(t)+np.random.randn(len(t))*0.1
     y=js.formel.smooth(data)  # 1d array
     #
     # smooth dataArray and replace .Y values.
     data2=js.dA(np.vstack([t,data]))
     data2.Y=js.formel.smooth(data2, windowlen=40, window='gaussian')

    Comparison of some filters:
    ::

     import jscatter as js
     import numpy as np
     t=np.r_[-5:5:0.01]
     data=js.dA(np.vstack([t,np.sin(t)+np.random.randn(len(t))*0.1]))
     p=js.grace()
     p.multi(4,2)
     windowlen=31
     for i,window in enumerate(['flat','gaussian','hann','fourier']):
         p[2*i].plot(data,sy=[1,0.1,6],le='original + noise')
         p[2*i].plot(t,js.formel.smooth(data,windowlen,window),sy=[2,0.1,4],le='filtered')
         p[2*i].plot(t,np.sin(t),li=[1,0.5,1],sy=0,le='noiseless')
         p[2*i+1].plot(data,sy=[1,0.1,6],le='original noise')
         p[2*i+1].plot(t,js.formel.smooth(data,windowlen,window),sy=[2,0.1,4],le=window)
         p[2*i+1].plot(t,np.sin(t),li=[1,2,1],sy=0,le='noiseless')
         p[2*i+1].text(window,x=-2.8,y=-1.2)
         p[2*i+1].xaxis(min=-3,max=-1,)
         p[2*i+1].yaxis(min=-1.5,max=-0.2,ticklabel=[None,None,None,'opposite'])
         p[2*i].yaxis(label='y')
     p[0].legend(x=10,y=4.5)
     p[6].xaxis(label='x')
     p[7].xaxis(label='x')
     p[0].title(f'Comparison of smoothing windows')
     p[0].subtitle(f'with windowlen {windowlen}')
     #p.save(js.examples.imagepath+'/smooth.jpg')


    .. image:: ../../examples/images/smooth.jpg
     :align: center
     :height: 300px
     :alt: smooth


    """
    if hasattr(data, '_isdataArray'):
        data = data.Y
    if window == 'flat':  # moving average
        window = 'boxcar'

    if window == 'fourier':
        # real fft; cut high frequencies; inverse fft
        rft = np.fft.rfft(np.r_[data[::-1], data])
        rft[int(2*len(data)/windowlen):] = 0
        smoothed = np.fft.irfft(rft)[data.size:]
        return smoothed

    if window in ['boxcar', 'flattop', 'hann', 'hamming', 'bartlett', 'blackman', 'gaussian']:
        windowlen = int(np.ceil(windowlen / 2) * 2)

        if data.size < windowlen:
            raise ValueError("Input vector needs to be bigger than window size.")

        if windowlen < 3:
            return data
        s = np.r_[data[windowlen - 1:0:-1], data, data[-1:-windowlen:-1]]

        if window == 'gaussian':  # gaussian
            window = ('gaussian', windowlen/7.)

        w = scipy.signal.get_window(window, windowlen)

        y = np.convolve(w / w.sum(), s, mode='valid')
        res = y[int((windowlen / 2 - 1)):int(-(windowlen / 2))]
        return res

    else:
        raise ValueError("Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman','gaussian'")




# -*- coding: utf-8 -*-
import numpy as np
import scipy.integrate.trapezoid as trapz
import jscatter as js

# read the data from Zetasizer exported data
# the wrapper takes the content from the header line which is consistent with the first data line
alldls = js.dls.readZetasizerNano(js.examples.datapath + '/dlsPolymerSolution.txt', NANOlocale=['de_DE', 'UTF-8'])
# we select one sample
# .X,.Y contain the correlation data
dls = alldls[4]

p = js.grace()
p.title('CONTIN analysis of DLS data')
p.multi(2, 1)
p[0].plot(dls, legend='correlation data')
p[0].plot(dls.distributions.X, dls.distributions.Y / 20, li=1, legend='relax. time intensity weighted')
p[1].plot(dls.distributions[1], dls.distributions[2] / 20, li=1, legend='radius intensity weighted')
# p[2].plot(dls.distributions[0],dls.distributions[2]/1000,li=1,legend='radius time intensity weighted')

p[0].xaxis(scale='l', label=r'correlation time / µs ', tick=[10, 9])
p[1].xaxis(scale='l', label=r'radius / nm ', tick=[10, 9])
# p[2].xaxis(scale='l',label=r'time / micros ',tick=[10,9])

# do the contin with the .X .Y data as intensity weighted relaxation time
resx = js.dls.contin(dls, a=173, tmin=1, tmax=1e5, typ=0, bgr=0.05, distribution='x', T=(273 + dls.Temperature))
p[0].plot(resx[0].contin_result_fit, li=4, sy=0, legend='contin result fit relaxation time')
p[0].plot(resx[0].contin_bestFit.X, resx[0].contin_bestFit.Y * 10, sy=[1, 0.3, 4], legend='relax. time distribution')

# determine the peak areas of the two peaks
ff = resx[0].contin_bestFit
p1 = (10 < ff.X) & (ff.X < 100)
fp1 = trapz(ff.Y[p1], ff.X[p1]) / trapz(ff.Y, ff.X)
p2 = (100 < ff.X) & (ff.X < 10000)
fp2 = trapz(ff.Y[p2], ff.X[p2])
p[0].text('fraction %.2g' % fp1, 90, 1.5, color=4)
p[0].text('fraction %.2g' % fp2, 10000, 0.5, color=4)


# do the contin with the .X .Y data as intensity weighted relaxation time
resd = js.dls.contin(dls, a=173, tmin=1, tmax=1e5, typ=0, bgr=0.05, distribution='d', T=(273 + dls.Temperature))

p[0].plot(resd[0].contin_result_fit, li=5, sy=0, legend='contin result fit d')
p[1].plot(resd[0].contin_bestFit.X, resd[0].contin_bestFit.Y * 1, legend='contin radius d')

# wavevector
# q=4*np.pi*1,333/632e-9*sin(theta/2)
# D=1/(G*q*q)
# Rh=k*T/(6*np.pi*visc*D)
# Rh=k*T*q*q/(6*np.pi*visc)  * G  = ff *G
kb = 1.3806505e-23  # J/K
theta = np.deg2rad(dls.ScatteringAngle)
q = 4 * np.pi * 1.333 / 632e-9 * np.sin(theta / 2)  # in 1/m
ff = kb * (dls.Temperature + 273.15) * q * q / (0.06 * np.pi * dls.Viscosity / 1000.)  # m/s
# Rh=ff*G
# .X time is in microseconds -> 1e-6 s
p[1].plot(resx[0].contin_bestFit.X * 1e-6 * ff * 1e9, resx[0].contin_bestFit.Y * weight, sy=[2, 0.3, 8],
          legend='relax. time to Rh')

p[0].legend(x=30000, y=2)
p[1].legend(x=0.01, y=15)

positions = []
print(resd.contin_bestFit[0].ipeaks_name)
for res in resd:
    isort = np.argsort(res.contin_bestFit.ipeaks[:, 1])
    sortedpeaks = res.contin_bestFit.ipeaks[isort]
    meanposition = np.r_[res.MeasurementTimeSeconds, sortedpeaks[:, 1]]
    positions.append(meanposition)

print('peak positions', positions)
p.save('Examplecontinanalysis.png')




"""
This test determines the best calphaCoarseGrainRadius to result in
best approximated forward scattering I0 of the selected proteins.

The best result is for calphaCoarseGrainRadius = 0.362

Dependent on the protein structure slightly different values might be better.
But anyway bead models and Cα models are always a rough approximation.

"""
import jscatter as js
import numpy as np
import os

with open(js.examples.datapath + '/proteinSpecificVolumes.txt') as f:
    lines = f.readlines()

vdw = {'CO': 0.15, 'Z': js.data.vdwradii['Zn'] * 10, 'M': js.data.vdwradii['Mg'] * 10}

Slist = js.dL()
for line in lines:
    words = line.split()
    if len(words)==0 or words[0].startswith('#'):
        continue
    if len(Slist)>0 and words[1] in Slist.proteinID:
        continue
    if os.path.exists(words[1]+'_h.pdb'):
        uni = js.bio.scatteringUniverse(words[1]+'_h.pdb', addHydrogen=False)
    else:
        uni = js.bio.scatteringUniverse(words[1])
    u = uni.select_atoms("protein")
    S = js.bio.xscatIntUniv(u)
    ca = uni.select_atoms('name CA')
    ca.write('uniCA.pdb')
    unica = js.bio.scatteringUniverse('uniCA.pdb', addHydrogen=False)
    uca = unica.residues
    pr = np.r_[0.3:0.4:0.01]
    for r in pr:
        print(words[1], r)
        unica.calphaCoarseGrainRadius = r
        Sca = js.bio.xscatIntUniv(uca)
        Sca.relVolume = Sca.SESVolume / S.SESVolume
        Sca.relI0 = Sca.I0 / S.I0
        Sca.atomVolume = S.SESVolume
        Sca.proteinID = words[1]
        Sca.pr = r
        Slist.append(Sca)

p2 = js.grace()
p2.multi(1, 2)
xx = []
for id in Slist.proteinID.unique:
    sl = Slist.filter(proteinID=id)
    rV = js.dA(np.c_[sl.pr, sl.relI0].T)
    rV.fit(lambda x, b, x0: (b * (x - x0) + 1) ** 2, {'b': 11, 'x0': 0.35}, {}, {'x': 'X'})
    p2[0].plot(rV, le=f'{rV.x0:.3f}')
    p2[0].plot(rV.lastfit, li=[1,0.3,11])
    p2[1].plot([sl[0].mass], [rV.x0])
    xx.append(rV.x0)
xx = np.array(xx)
p2[0].plot([0.1, 0.4], [1] * 2, sy=0, li=1)
p2[0].xaxis(label='Ca radius / nm', min=0.28, max=0.45)
p2[0].yaxis(label='relative intensity')
p2[1].xaxis(scale='log', label='mass / Da')
p2[1].yaxis(label=['best Ca radius / nm',1,'opposite'],ticklabel= ['general',2,1,'opposite'])
p2[1].subtitle(f'best Ca CG radius = {np.mean(xx):.3f} +- {np.std(xx):.3f} ')
p2[0].title('                            Determination of best Ca coarse grain radius')
# p2.save('proteinCacoarsegrainRadius.png')

# Conclusion CA model with vdWradius 0.342 is appropriate




# Basic fit examples with synthetic data. Usually data are loaded from a file.

import jscatter as js
import numpy as np

sinus = lambda x, A, a, B: A * np.sin(a * x) + B  # define model

# Fit sine to simulated data
x = np.r_[0:10:0.1]
data = js.dA(np.c_[x, np.sin(x) + 0.2 * np.random.randn(len(x)), x * 0 + 0.2].T)  # simulate data with error
data.fit(sinus, {'A': 1.2, 'a': 1.2, 'B': 0}, {}, {'x': 'X'})  # fit data
data.showlastErrPlot()  # show fit
data.errPlotTitle('Fit Sine')

# Fit sine to simulated data using an attribute in data with same name
data = js.dA(np.c_[x, 1.234 * np.sin(x) + 0.1 * np.random.randn(len(x)), x * 0 + 0.1].T)  # create data
data.A = 1.234  # add attribute
data.makeErrPlot()  # makes errorPlot prior to fit
data.fit(sinus, {'a': 1.2, 'B': 0}, {}, {'x': 'X'})  # fit using .A
data.errPlotTitle('Fit Sine with attribute')

# Fit sine to simulated data using an attribute in data with different name and fixed B
data = js.dA(np.c_[x, 1.234 * np.sin(x) + 0.1 * np.random.randn(len(x)), x * 0 + 0.1].T)  # create data
data.dd = 1.234  # add attribute
data.fit(sinus, {'a': 1.2, }, {'B': 0}, {'x': 'X', 'A': 'dd'})  # fit data
data.showlastErrPlot()  # show fit
data.errPlotTitle('Fit Sine with attribute and fixed B')

# Fit sine to simulated dataList using an attribute in data with different name
# and fixed B from data.
# first one common parameter then as parameter list
# create data
data = js.dL()
ef = 0.1  # increase this to increase error bars of final result
for ff in [0.001, 0.4, 0.8, 1.2, 1.6]:
    data.append(js.dA(np.c_[x, (1.234 + ff) * np.sin(x + ff) + ef * ff * np.random.randn(len(x)), x * 0 + ef * ff].T))
    data[-1].B = 0.2 * ff / 2  # add attributes

# fit with a single parameter for all data, obviously wrong result
data.fit(lambda x, A, a, B, p: A * np.sin(a * x + p) + B, {'a': 1.2, 'p': 0, 'A': 1.2}, {}, {'x': 'X'})
data.showlastErrPlot()  # show fit
data.errPlotTitle('Fit Sine with attribute and common fit parameter')

# now allowing multiple p,A,B as indicated by the list starting value
data.fit(lambda x, A, a, B, p: A * np.sin(a * x + p) + B, {'a': 1.2, 'p': [0], 'B': [0, 0.1], 'A': [1]}, {}, {'x': 'X'})
data.errPlotTitle('Fit Sine with attribute and non common fit parameter')

# plot p against A , just as demonstration
p = js.grace()
p.plot(data.A, data.p, data.p_err, sy=[1, 0.3, 1])
p.xaxis(label='Amplitude')
p.yaxis(label='phase')



# Fit sine to simulated dataList with a linked parameter
#
import jscatter as js
import numpy as np

# create data. B will appear several times with same value
data = js.dL()
x = np.r_[0:10:0.1]
ef = 0.1  # increase this to increase error bars of final result
for ff in [0.001, 0.8, 1.6]:
    data.append(js.dA(np.c_[x, (1.234 + ff) * np.sin(x + ff) + ef * ff * np.random.randn(len(x)), x * 0 + ef * ff].T))
    data[-1].B = 0.2 * ff / 2  # add attributes
    # add same but different amplitude A, same phase p
    data.append(js.dA(np.c_[x, (0.234 + ff) * np.sin(x + ff) + ef * ff * np.random.randn(len(x)), x * 0 + ef * ff].T))
    data[-1].B = 0.2 * ff / 2  # add attributes

data.append(js.dA(np.c_[x, (0.234 + ff) * np.sin(x + ff) + ef * ff * np.random.randn(len(x)), x * 0 + ef * ff].T))
data[-1].B = 0.2 * ff / 2  # add attributes


# create model
model = lambda x, A, a, B, p: A * np.sin(a * x + p) + B

# plot for result
p = js.grace()

# fit with independent parameters for all dataList elements
data.fit(model, {'a': 1.2, 'p': [0], 'A': [1.2]}, {}, {'x': 'X'})

# plot p against unique B
p.plot(data.B, data.p, data.p_err, sy=[1, 1, 1], legend='independent p with 7 values')

# now A still independent but we link 'p' to 'B'
data.fit(model, {'a': 1.2, 'p': [0, 'B'], 'A': [1]}, {}, {'x': 'X'})

# plot p against unique B with less values in p
p.plot(data.B.unique, data.p, data.p_err, sy=[2, 0.7, 4], le='linked p with 3 values')
p.xaxis(label='B', min=-0.1, max=0.25)
p.yaxis(label='phase',min=-0.5,max=2.5)
p.legend(x=-0.05,y=2)
p.save('SinusoidalFit_linkedParameter.png')




cdef extern from './cpackage/cubature.h':
    # values for the error_norm parameter
    ctypedef enum error_norm:
        ERROR_INDIVIDUAL = 0
        ERROR_PAIRED
        ERROR_L2
        ERROR_L1
        ERROR_LINF

    ctypedef int (*integrand) (unsigned ndim, const double *x, void *fdata,
                               unsigned fdim, double *fval)

    ctypedef int (*integrand_v) (unsigned ndim, size_t npt, const double *x,
                                 void *fdata, unsigned fdim, double *fval)

    int hcubature(unsigned fdim, integrand f, void *fdata,
                  unsigned ndim, const double *xmin, const double *xmax,
                  unsigned maxEval, double reqAbsError, double reqRelError,
                  error_norm norm, double *val, double *err)

    int pcubature(unsigned fdim, integrand f, void *fdata,
                  unsigned ndim, const double *xmin, const double *xmax,
                  size_t maxEval, double reqAbsError, double reqRelError,
                  error_norm norm, double *val, double *err)

    int hcubature_v(unsigned fdim, integrand_v f, void *fdata,
                    unsigned ndim, const double *xmin, const double *xmax,
                    size_t maxEval, double reqAbsError, double reqRelError,
                    error_norm norm, double *val, double *err)

    int pcubature_v(unsigned fdim, integrand_v f, void *fdata,
                    unsigned ndim, const double *xmin, const double *xmax,
                    size_t maxEval, double reqAbsError, double reqRelError,
                    error_norm norm, double *val, double *err)
# Vectorized version with user-supplied buffer to store points and values.
# The buffer *buf should be of length *nbuf * dim on entry (these parameters
# are changed upon return to the final buffer and length that was used).
# The buffer length will be kept <= max(max_nbuf, 1) * dim.
#
# Also allows the caller to specify an array m[dim] of starting degrees
# for the rule, which upon return will hold the final degrees.  The
# number of points in each dimension i is 2^(m[i]+1) + 1.
    int pcubature_v_buf(unsigned fdim, integrand_v f, void *fdata,
                    unsigned ndim, const double *xmin, const double *xmax,
                    size_t maxEval, double reqAbsError, double reqRelError,
                    error_norm norm, unsigned *m,
                    double **buf, size_t *nbuf, size_t max_nbuf,
                    double *val, double *err)




#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=True
#cython: nonecheck=False
#cython: infer_types=False

cimport numpy as np
import numpy as np
import cython
from cpython.ref cimport PyObject
from ._cubature cimport (error_norm, integrand, integrand_v, hcubature, pcubature,
        hcubature_v, pcubature_v)

cdef extern from "get_ptr.h":
    void *get_ctypes_function_pointer(PyObject *obj)

cdef class Integrand:
    cdef object f, args, kwargs
    cdef unsigned int ndim, fdim

    def __cinit__(self, object f, unsigned ndim, unsigned fdim, object args,
            object kwargs):
        self.f = f
        self.ndim = ndim
        self.fdim = fdim
        self.args = args
        self.kwargs = kwargs

    def __init__(self, *args, **kwargs):
        if not callable(self.f):
            raise ValueError('first argument not callable')

    def __str__(self):
        s = 'Integrand(f = {!r}, ndim = {!r}, fdim = {!r}, args = {!r}, kwargs = {!r})'\
             .format(self.f, self.ndim, self.fdim, self.args, self.kwargs)
        return s

    cdef int _call(self, const double *x, double *fval) except -1:
        cdef double [:] _x = <double [:self.ndim]>x
        cdef double [:] _f = <double [:self.fdim]>fval
        cdef int error
        cdef unsigned i

        try:
            tmp = self.f(np.asarray(_x), *self.args, **self.kwargs)
            if self.fdim == 1:
                _f[0] = tmp
            else:
                for i in range(self.fdim):
                    _f[i] = tmp[i]
            error = 0
        except Exception as e:
            error = -1
            raise e
        return error

    cdef int _vcall(self, unsigned npts, const double *x, double *fval) except -1:
        cdef double [:, :] _x = <double [:npts, :self.ndim]>x
        cdef double [:, :] _f = <double [:npts, :self.fdim]>fval
        cdef int error
        cdef unsigned i,j

        try:
            tmp = self.f(np.asarray(_x), *self.args, **self.kwargs)
            if self.fdim == 1:
                for i in range(npts):
                    _f[i] = tmp[i]
            else:
                for i in range(npts):
                    for j in range(self.fdim):
                        _f[i,j] = tmp[i,j]
            error = 0
        except Exception as e:
            error = -1
            raise e
        return error

    def call(self, xval):
        cdef double [:] _xval = np.array(xval, dtype=np.float64)
        cdef double [:] fval = np.zeros((self.fdim,), dtype=np.float64)
        err = self._call(&_xval[0], &fval[0])
        if err != 0:
            raise RuntimeError('error while calling call()')
        return np.asarray(fval)

    def vcall(self, xval):
        cdef double [:, :] _xval = np.array(xval, dtype=np.float64)
        cdef unsigned npts = _xval.shape[0]
        cdef double [:, :] fval = np.zeros((npts, self.fdim,), dtype=np.float64)
        err = self._vcall(npts, &_xval[0,0], &fval[0,0])
        if err != 0:
            raise RuntimeError('error while calling vcall()')
        return np.asarray(fval)


cdef int integrand_wrapper(unsigned int ndim, double *x, void *fdata,
        unsigned int fdim, double *fval):
    cdef Integrand wrapped = <Integrand>fdata;
    return wrapped._call(x, fval)


cdef int integrand_wrapper_v(unsigned int ndim, unsigned int npts, double *x,
        void *fdata, unsigned int fdim, double *fval):
    cdef Integrand wrapped = <Integrand>fdata;
    return wrapped._vcall(npts, x, fval)


def cubature(callable, unsigned ndim, unsigned fdim, xmin, xmax, str method,
        double abserr, double relerr, int norm, unsigned maxEval, args=(),
        kwargs={}):

    cdef double [:] _xmin = np.array(xmin, dtype=np.float64)
    cdef double [:] _xmax = np.array(xmax, dtype=np.float64)

    cdef double [:] val = np.empty((fdim,), dtype=np.float64)
    cdef double [:] err = np.empty((fdim,), dtype=np.float64)

    wrapper = Integrand(callable, ndim, fdim, args, kwargs)

    if method == 'hcubature_v':
        error = hcubature_v(fdim, <integrand_v>integrand_wrapper_v,
                <void *> wrapper, ndim, &_xmin[0], &_xmax[0], maxEval, abserr,
                relerr, <error_norm> norm, &val[0], &err[0])

    elif method == 'hcubature':
        error = hcubature(fdim, <integrand>integrand_wrapper, <void *> wrapper,
                ndim, &_xmin[0], &_xmax[0], maxEval, abserr, relerr,
                <error_norm> norm, &val[0], &err[0])

    elif method == 'pcubature_v':
        error = pcubature_v(fdim, <integrand_v>integrand_wrapper_v,
                <void *> wrapper, ndim, &_xmin[0], &_xmax[0], maxEval, abserr,
                relerr, <error_norm> norm, &val[0], &err[0])

    elif method == 'pcubature':
        error = pcubature(fdim, <integrand>integrand_wrapper, <void *> wrapper,
                ndim, &_xmin[0], &_xmax[0], maxEval, abserr, relerr,
                <error_norm> norm, &val[0], &err[0])

    else:
        raise ValueError('unknown integration method `{!s}`'.format(method))

    if error != 0:
        raise RuntimeError('integration failed')

    return np.asarray(val), np.asarray(err)

def cubature_raw_callback(callable, unsigned ndim, unsigned fdim, xmin, xmax, str method,
        double abserr, double relerr, int norm, unsigned maxEval, args=(),
        kwargs={}):

    cdef double [:] _xmin = np.array(xmin, dtype=np.float64)
    cdef double [:] _xmax = np.array(xmax, dtype=np.float64)

    cdef double [:] val = np.empty((fdim,), dtype=np.float64)
    cdef double [:] err = np.empty((fdim,), dtype=np.float64)

    if method == 'hcubature_v':
        error = hcubature_v(fdim, <integrand_v>get_ctypes_function_pointer(<PyObject *>callable),
                NULL, ndim, &_xmin[0], &_xmax[0], maxEval, abserr,
                relerr, <error_norm> norm, &val[0], &err[0])

    elif method == 'hcubature':
        error = hcubature(fdim, <integrand>get_ctypes_function_pointer(<PyObject *>callable), NULL,
                ndim, &_xmin[0], &_xmax[0], maxEval, abserr, relerr,
                <error_norm> norm, &val[0], &err[0])

    elif method == 'pcubature_v':
        error = pcubature_v(fdim, <integrand_v>get_ctypes_function_pointer(<PyObject *>callable),
                NULL, ndim, &_xmin[0], &_xmax[0], maxEval, abserr,
                relerr, <error_norm> norm, &val[0], &err[0])

    elif method == 'pcubature':
        error = pcubature(fdim, <integrand>get_ctypes_function_pointer(<PyObject *>callable), NULL,
                ndim, &_xmin[0], &_xmax[0], maxEval, abserr, relerr,
                <error_norm> norm, &val[0], &err[0])

    else:
        raise ValueError('unknown integration method `{!s}`'.format(method))

    if error != 0:
        raise RuntimeError('integration failed')

    return np.asarray(val), np.asarray(err)



#
# Copyright 2000 by Peter McCluskey (pcm@rahul.net).
# You may do anything you want with it, provided this notice is kept intact.
#
# Changed to python 3 compatibility and usage of MdAnalysis Ralf Biehl 2021
# You may do anything you want with it, provided this notice is kept intact.

"""
_surface_atoms(atoms, solvent_radius = 0., point_density = 258, ret_fmt = 0)

point_density must be 2**(2*N) + 2, where N > 1 (258 and 1026 seem best).
Returns a dictionary with an item for each input atom.
These choices for ret_fmt specify what the dictionary will hold for each atom:
 1: area
 2: (area, volume)
 3: (area, volume, points)
 4: (area, volume, points, dir_tuple)

 points is a list of 3-tuples describing coordinates of points on a
solvent-accessible surface (zero to point_density points per atom).

 dir_tuple is a 3-tuple giving crude estimate of the direction which is
locally "up", i.e. normal to the surface. It is calculated by comparing
the atom's position with the average position of an atom's accessible surface
points.

 The algorithm used is based on the method described in this paper:
 Eisenhaber, Frank, et al. "The Double Cubic Lattice Method: Efficient
Approaches to Numerical Integration of Surface Area and Volume and to
Dot Surface Contouring of Molecular Assemblies", Journal of Computational
Chemistry, Vol. 16, pp 273-284 (1995).
 I have taken a few shortcuts which probably make it a bit less accurate
than what the paper describes (in particular, I used a simple tesselation
algorithm without fully understanding the one described in the paper).

"""

import math
import numpy as np
import multiprocessing as mp

try:
    from ..libs import surface
    useC=True
except:
    print('use python in SASAsurface.py')
    useC = False

# a global object that contains neighbor class to avoid pickling
# this speeds up using a pool of workers. Should be safe as only one fit at a time is running.
_pooldata = None

"""
Surface related functions

"""


def aVSPointsGradients(objekt, probe_radius=0.13, shellthickness=0.3, point_density=4, vdwradii=None):
    """
    Calcs the solvent accessible surface by the rolling ball algorithm.

    Parameters
    ----------
    objekt :
        Object with atoms. Atoms need method .type or .name to get key in vdwradii (capital letters like 'C')
    probe_radius : float
        Distance from the vdW-radii of the atoms at which the surface is computed.
    shellthickness : float
        Thickness of solvent shell for shellvolume.
    point_density : int
        Density of points that describe the surface.
        The point density is finally 2**(2*point_density)+2 => (4 -> 258; 5 -> 1026)
    vdwradii : dict, default None
        Dictionary of {element symbol : van der Waals radii in nm}
        as {'C': 0.17,'H': 0.12,'N': 0.155,'O': 0.152,'S':0.18}

    Returns
    -------
        [total surface area, total surface layer volume, surface atom dict]
        The surface atom dict maps surface atoms to a tuple containing 5 surface-related quantities per surface atom:
         - SAS exposed surface area
         - list of points in the exposed surface
         - gradient vector pointing outward from the surface.
         - number of points at proberadius/2
         - number of points at proberadius
         - number of points at shellthickness

    """
    atoms = objekt.atoms
    smap = _surface_atoms(atoms,
                          probe_radius,
                          shellthickness,
                          point_density=2 ** (2 * point_density) + 2,
                          vdwradii=vdwradii)
    surface_data = {}
    tot_a = 0
    tot_v = 0
    for a in atoms:
        (area, volume, points1, grad, lenr2, lenr, lenshell) = smap[a]
        if lenr2 > 0.:
            # we have a surface atom at least in probe_radius/2
            surface_data[a] = (points1, area, grad, lenr2, lenr, lenshell)
            tot_a = tot_a + area
            tot_v = tot_v + volume
    return tot_a, tot_v, surface_data


def _get_atom_data(atoms, vdwradii):
    """
    Returns tuple with max vdW radii, atom_data, geometric center.

    Parameters
    ----------
    atoms : atom objects
        atoms with positionattribut. For mda use posnm as position in nm
    vdwradii : vdW radii in units nm

    Returns
    -------

    """
    atom_data = [None] * len(atoms)
    maxvdWrad = 0
    sumx = 0
    sumy = 0
    sumz = 0
    for i, a in enumerate(atoms):
        try:
            # priority for attributed values
            vdw = a.vdWradiinm
        except (KeyError, AttributeError):
            vdw = vdwradii[a.type.upper()]
        pos1 = a.posnm  # conversion to nm from mda in A
        atom_data[i] = (pos1[0], pos1[1], pos1[2], vdw)
        sumx = sumx + atom_data[i][0]
        sumy = sumy + atom_data[i][1]
        sumz = sumz + atom_data[i][2]
        maxvdWrad = max(maxvdWrad, atom_data[i][3])

    return maxvdWrad, atom_data, (sumx/len(atoms), sumy/len(atoms), sumz/len(atoms))


def _surface_atoms(atoms, solvent_radius=0., shellthickness=0, point_density=258, vdwradii=None):
    global _pooldata

    (maxrad, atom_data, center) = _get_atom_data(atoms, vdwradii)

    tess1 = None

    # create neighbor list for fast search
    # and put all to global data to avoid pickling in pool
    _pooldata = {'point_density': point_density,
                 'center': center,
                 'neighbors': NeighborList(atoms,  maxrad + max(solvent_radius, shellthickness), atom_data),
                 'atom_data': atom_data,
                 'tess1': tess1,
                 'solvent_radius': solvent_radius,
                 'shellthickness': shellthickness}

    if 'fork' in mp.get_all_start_methods():
        with mp.get_context('fork').Pool() as pool:
            chunksize, extra = divmod(len(atoms), len(pool._pool) * 4)
            chunksize = chunksize + (1 if extra else 0)
            result = pool.map_async(_asurf, range(len(atoms)), chunksize=chunksize).get()
    else:
        # just in case linux and MacOs change support and in case Windows works
        with mp.pool.ThreadPool() as pool:
            chunksize, extra = divmod(len(atoms), len(pool._pool) * 4)
            chunksize = chunksize + (1 if extra else 0)
            result = pool.map_async(_asurf, range(len(atoms)), chunksize=chunksize).get()

    surf_points = {atoms[i]: res for i, res in result}

    # reset pooldata
    _pooldata = None

    return surf_points


def _asurf(i):
    ret_fmt = 4

    # get data from global pooldata
    pos1 = _pooldata['atom_data'][i]
    vdWrad = pos1[3]
    neighbors = _pooldata['neighbors']
    atom_data = _pooldata['atom_data']
    point_density = _pooldata['point_density']
    tess1 = _pooldata['tess1']
    center = _pooldata['center']
    solvent_radius = _pooldata['solvent_radius']
    shellthickness = _pooldata['shellthickness']

    # first in probe_radius/2 to get best corner estimate and maximal amount of surface atoms
    (halfpoints1, halfpoints_unit) = _atom_surf(neighbors, i, atom_data, pos1, solvent_radius/2,
                                        point_density, tess1, False)
    if halfpoints1:
        # only if atom is surface atom (halfpoints1 is True)
        # probe_radius -> most important to get points on surface
        (points1, points_unit) = _atom_surf(neighbors, i, atom_data, pos1, solvent_radius,
                                            point_density, tess1, True)
        (shell_points1, shell_points_unit) = _atom_surf(neighbors, i, atom_data, pos1, shellthickness,
                                            point_density, tess1, False)
        if points1:
            res = _xlate_results(points1, points_unit, point_density, vdWrad+solvent_radius, pos1, ret_fmt, center)
            return i, res + (len(halfpoints1), len(points1), len(shell_points1))
        else:
            return i, (0, 0, 0, 0, len(halfpoints1), 0, len(shell_points1))

    else:
        return i, (0, 0, 0, 0, 0, 0, 0)


def _atom_surf(nbors, i, atom_data, pos1, proberad, point_density, tess1, ret_unit_points=1):
    if useC:
        return surface.surface1atom(nbors, i, atom_data, pos1, proberad, point_density, ret_unit_points)

    else:
        # the slow python version of the C code for ret_unit_points=1
        tot_rad = atom_data[i][3] + proberad
        rad1sq = tot_rad*tot_rad
        rad1_2 = 2*tot_rad
        data = []
        for (index, d2) in nbors[i]:
            apos = atom_data[index]
            vaax = apos[0] - pos1[0]
            vaay = apos[1] - pos1[1]
            vaaz = apos[2] - pos1[2]
            thresh = (d2 + rad1sq - (apos[3]+proberad)**2) / rad1_2
            data.append((vaax, vaay, vaaz, thresh))
        points_unit = []  # points on unit sphere
        points1 = []  # points on sphere radius = tot_rad
        for pt1 in tess1:
            buried = 0
            for (vx, vy, vz, thresh) in data:
                if vx*pt1[0] + vy*pt1[1] + vz*pt1[2] > thresh:
                    buried = 1
                    break
            if buried == 0:
                points1.append((tot_rad*pt1[0] + pos1[0],
                                tot_rad*pt1[1] + pos1[1],
                                tot_rad*pt1[2] + pos1[2]))
                if ret_unit_points:
                    points_unit.append(pt1)
        return points1, points_unit


class NeighborList:

    """
    This class is designed to efficiently find lists of atoms which are
    within a distance "radius" of each other.

    Constructor: NeighborList(|atoms|, |radius|, |atom_data|)

    Arguments:

    |atoms| - list of MMTKnp Atom
    |radius| - max distance between neighboring atoms
    |atom_data| - data returned from _get_atom_data
    """

    def __init__(self, atoms, radius, atom_data):
        boxes = {}
        self.box_size = 2*radius
        for i in range(len(atoms)):
            f = self.box_size
            pos = atom_data[i]
            key = '%d %d %d' % (int(math.floor(pos[0]/f)),
                                int(math.floor(pos[1]/f)),
                                int(math.floor(pos[2]/f)))
            if key in boxes:
                boxes[key].append(i)
            else:
                boxes[key] = [i]
        self.boxes = boxes
        self.atoms = atoms
        self.atom_data = atom_data

    def __len__(self):
        return len(self.atoms)

    def __getitem__(self, i):
        """
        Returns a list of tuples describing the neighbors of the ith atom
        in the atom list. Each tuple has the index of the atom which neighbors
        atom i, followed by the square of the distance between atoms.
        """
        boxes = self.boxes
        if useC:
            return surface.FindNeighborsOfAtom(self.atoms, i, boxes, self.box_size, self.atom_data)

        max_dist_2 = self.box_size**2
        pos1 = self.atom_data[i]
        f = self.box_size
        key_tup = (int(math.floor(pos1[0]/f)),
                   int(math.floor(pos1[1]/f)),
                   int(math.floor(pos1[2]/f)))
        nlist = []
        for x in (-1, 0, 1):
            for y in (-1, 0, 1):
                for z in (-1, 0, 1):
                    key2 = '%d %d %d' % (key_tup[0]+x, key_tup[1]+y, key_tup[2]+z)
                    if key2 in boxes:
                        for i2 in boxes[key2]:
                            if i2 != i:
                                apos = self.atom_data[i2]
                                vaax = apos[0] - pos1[0]
                                vaay = apos[1] - pos1[1]
                                vaaz = apos[2] - pos1[2]
                                d2 = vaax*vaax + vaay*vaay + vaaz*vaaz
                                if d2 > max_dist_2:
                                    continue
                                nlist.append((i2, d2))
        return nlist


def _xlate_results(surfpoints, points_unit, point_density, tot_rad, pos1, ret_fmt, center):
    area = 4 * math.pi * (tot_rad**2) * len(surfpoints) / point_density

    if ret_fmt < 2:
        return area

    if points_unit:
        sumx, sumy, sumz = np.sum(points_unit, axis=0)
        n = max(1, len(surfpoints))
        vconst = 4/3.0*math.pi/point_density
        dotp1 = (pos1[0] - center[0])*sumx + (pos1[1] - center[1])*sumy + (pos1[2] - center[2])*sumz
        volume = vconst*((tot_rad**2) * dotp1 + (tot_rad**3) * len(surfpoints))
        surfp = np.array(surfpoints, dtype=np.float32)
        grad = (sumx/n, sumy/n, sumz/n)
    else:
        # points_unit == []
        volume = 0
        surfp = np.array(surfpoints, dtype=np.float32)
        grad = (0, 0, 0)

    if ret_fmt == 2:
        return area, volume
    elif ret_fmt == 4:
        return area, volume, surfp, grad
    else:
        return area, volume, surfp





""" Alcohol dehydrogenase (yeast) example for normal mode analysis
#
# See Biehl et al.PRL 101, 138102 (2008)

In this example we consecutivly
 - load the protein
 - create normal modes
 - calc effective diffusion
 - calc the dynamic mode formfactor from normal modes
 - show how the intermediate scattering function looks with diffusion and internal normal mode relaxation

Finally, we build from this a model that can be used for fitting

"""

# create universe with protein inside that adds hydrogen
#  - load PDB structure
#  - repair structure e.g. missing atoms 
#  - add hydrogen using pdb2pqr, saving this to 3cln_h.pdb
#  - adds scattering amplitudes, volume determination

%matplotlib
import jscatter as js
import numpy as np

adh = js.bio.fetch_pdb('4w6z.pdb1')
# the 2 dimers in are in model 1 and 2 and need to be merged into one.
adhmerged = js.bio.mergePDBModel(adh)
u = js.bio.scatteringUniverse(adhmerged)
u.atoms.deuteration = 0
protein = u.select_atoms("protein")
S0 = js.bio.nscatIntUniv(protein, cubesize=0.5, getVolume='once')

# ## Calc effective diffusion with trans/rot contributions
# - determine $D_{trans}$ and $D_{rot}$ using HULLRAD
# - calc Deff
D_hr = js.bio.hullRad(u)
Dt = D_hr['Dt'] * 1e2  # conversion to nm²/ps
Dr = D_hr['Dr'] * 1e-12  # conversion to 1/ps
u.qlist = np.r_[0.01, 0.1:3:0.2]
Deff = js.bio.diffusionTRUnivTensor(u.residues, DTT=Dt, DRR=Dr, cubesize=0.5)

p=js.mplot()
p.Plot(Deff.X, Deff.Y*1e5, li=1, le='rigid ADH protein')
p.Xaxis(label='$Q / nm^{-1}$')
p.Yaxis(label='$D_{eff} /A^2/ns$')

# ## Create normal modes based on residues
nm = js.bio.vibNM(protein.residues,k_b=418)

# ### Normal mode relaxation in small displacement approximation
Ph678 = js.dL()
for NN in [6,7,8]:
   Ph = js.bio.intScatFuncPMode(nm, NN, output=0, qlist=Deff.X)
   Ph678.append(Ph)

# ## effective diffusion Deff in initial slope (compare to cumulant fit)
a=100.
rate = 1/30000 # 1/ps
for Ph in Ph678:
   d = Deff.interp(Ph.X) + rate * a**2 * Ph._Pn / (Ph._Fq+a**2*Ph._Pn) / Ph.X**2
   p.Plot(Ph.X,1e5*d ,li='', le=f'rigid ADH + mode {Ph.modeNumber}  rmsd={Ph.kTrmsd*a:.2f} A')

p.Title('Alcohol dehydrogenase (ADH) effective diffusion \nwith additional normal mode relaxations')
p.Legend(x=1.5,y=5.5)
# p.savefig(js.examples.imagepath+'/ADHNM_Deff.jpg', dpi=100)

# Assume a common relaxation on top of diffusion
# that we add to  Deff
u.qlist = np.r_[0.2:2:0.2]    # [1/nm]
u.tlist = np.r_[1, 10:1e5:50]  # [ps]
Iqt = js.bio.intScatFuncYlm(u.residues,Dtrans=Dt,Drot=Dr,cubesize=1,getVolume='once')

# ### dynamic mode formfactor P() and relaxation in small displacement approximation with amplitude A(Q)
sumP = Ph678.Y.array.sum(axis=0)
def Aq(a):
    # NM mode formfactor amplitude sum
    aq = a**2*sumP / (Ph678[0]._Fq + a**2*sumP)
    return js.dA(np.c_[Ph678[0].X, aq].T)

p2=js.mplot()
p2.Yaxis(min=0.01, max=1, label='I(Q,t)/I(Q,0)')
p2.Xaxis(min=0, max=100000, label='t / ps')

Iqt2 = Iqt.copy()
l=1/10000  # 1/ps
Aqa = Aq(a)
for i, qt in enumerate(Iqt2):
    diff = qt.Y *(1-Aqa.interp(qt.q))
    qt.Y = qt.Y *((1-Aqa.interp(qt.q)) + Aqa.interp(qt.q)*np.exp(-l*qt.X))
     
    p2.Plot(qt.X, qt.Y * 0.8**i,sy=0,li=[3,2,i+1],le=f'{qt.q:.1f}')
    p2.Plot(qt.X, diff* 0.8**i,sy=0,li=[1,2,i+1])
    
p2.Yaxis(min=0.001,max=1,label='I(Q,t)/I(Q,0)',scale='log')
p2.Xaxis(min=0.1,max=100000,label='t / ps')
p2.Title('Intermediate scattering function with/without NM relaxation')
p2.Subtitle('scaled for visibility')
p2.Legend()
# p2.savefig(js.examples.imagepath+'/ADHNM_Iqt.jpg', dpi=100)

if 0 :
    # look at A(Q)
    p1=js.mplot()
    Aqa = Aq(a)
    p1.Plot(Aqa, sy=0, li=1)
    p1.Yaxis(min=0, max=0.8)




# in nutshell without fine tuning of plots

import jscatter as js
import numpy as np

# read data with 16 intermediate scattering functions from NSE measurement of protein diffusion
i5 = js.dL(js.examples.datapath + '/iqt_1hho.dat')

# manipulate data
for dat in i5:
    dat.X = dat.X /1.  # conversion from ps to ns
    dat.q *= 1  # conversion to 1/nm if needed

# define model as simple diffusion with elastic background
diffusion = lambda A, D, t, elastic, wavevector=0: A * np.exp(-wavevector ** 2 * D * t) + elastic

# make ErrPlot to see progress of intermediate steps with residuals (updated all 2s)
i5.makeErrPlot(title='diffusion model residual plot', legpos=[0.2, 0.8])

# fit it
i5.fit(model=diffusion,  # the fit function
       freepar={'D': [0.2, 0.25], 'A': 1},  # fit start values; [..] indicate independent fit parameter
       fixpar={'elastic': 0.0},             # single values indicates common fit parameter
       mapNames={'t': 'X', 'wavevector': 'q'},  # map names of the model to names of data attributes
       condition=lambda a: (a.X > 0.01) & (a.Y > 0.01))  # a condition to include only specific values

# save fit result with errors and covariance matrix
i5.lastfit.savetxt('iqt_proteindiffusion_fit.dat')

# plot it together with lastfit result
p = js.grace()
p.plot(i5, symbol=[-1, 0.4, -1], legend='Q=$q')  # plot as alternating symbols and colors with size 0.4
p.plot(i5.lastfit, symbol=0, line=[1, 1, -1])  # plot a line with alternating colors
p.save('iqt_proteindiffusion_fit.png')

# plot result with error bars
p1 = js.grace(2, 2)  # plot with a defined size
p1.plot(i5.lastfit.wavevector, i5.lastfit.D, i5.lastfit.D_err, symbol=[2, 1, 1, ''], legend='average effective D')
p1.save('Diffusioncoefficients.agr')  # save as XmGrace plot



import jscatter as js
import numpy as np
import random
gp=js.graceplot

x = np.r_[1:11]
y = x
dy = map(lambda x: random.random() * 2., x)
labels = ['pt1', 'pt2', 'Iridium', 'Na', 'Ti', 'hydrogen', 'Mo ' + "1.23e3", 'Ta',
          'pokemon',
          'digital']

p = js.grace(width=2, height=1.5)  # A grace session opens

# original idea to create symbols and lines and data object
s1 = gp.Symbol(symbol=gp.symbols.square, fillcolor=gp.colors.cyan)
l1 = gp.Line(type=gp.lines.none)

d1 = gp.DataXYDY(x=x, y=y, dy=dy, symbol=s1, line=l1)

g = p[0]
g.xaxis(min=0, max=12)
g.yaxis(min=0, max=12)

g.plot(d1, autoscale=False)

for i in range(len(labels)):
    g.text('  ' + labels[i], x[i], y[i], color=gp.colors.violet, charsize=1.2)

g.line(x1=3, y1=1, x2=8, y2=2, linewidth=3, color=gp.colors.green4)
p.save('grace3.png')



import jscatter as js
import numpy as np
import scipy.constants as constants

# create universe
mab = js.bio.scatteringUniverse('1igt')
mab.setSolvent('1d2o1')

mab.qlist = np.r_[0.01, 0.1:4:0.03]
# use explicit residue formfactors for EACH residue
mab.explicitResidueFormFactorAmpl = True

# use a atom selection
lightchains = mab.select_atoms('segid A C')  # light chains

mab.atoms.deuteration = 0  # default protonated
I_protonated = js.bio.nscatIntUniv(mab.atoms)

lightchains.atoms.deuteration = 1
I_lightdeuterated = js.bio.nscatIntUniv(mab.atoms)

mab.atoms.deuteration = 0
# select fc domain
mab.select_atoms('segid B D and resid 240-700').deuteration = 1
I_fcdeuterated = js.bio.nscatIntUniv(mab.atoms)

mab.atoms.deuteration = 0
# select 3 residue
mab.select_atoms('resname ALA ARG ASN ').deuteration = 1
I_someAA = js.bio.nscatIntUniv(mab.atoms)
#

I_someresidue = js.bio.nscatIntUniv(mab.residues)

p = js.grace()
p[0].plot(I_protonated, sy=0, li=[1, 3, 1], le='full protonated')
p[0].plot(I_lightdeuterated, sy=0, li=[1, 3, 2], le='light chain deuterated')
p[0].plot(I_fcdeuterated, sy=0, li=[1, 3, 3], le='fc domain deuterated')
p[0].plot(I_someAA, sy=0, li=[1, 3, 4], le='ALA ARG ASN deuterated')
p[0].plot(I_someresidue, sy=0, li=[3, 3, 5], le='ALA ARG ASN deuterated coarse grain')


p.xaxis(label=r'Q / nm\S-1', min=0, max=4)
p.yaxis(label=r'I(Q) / 1/cm', scale='log', min=1e-6, max=0.01)
p.legend(x=1, y=0.008)
p.title('selective protein deuteration')

# p.save('bio_protein_partialmatching.png')

